#ifndef COMMON_FUNCTIONS_H
#define COMMON_FUNCTIONS_H

#include <iostream>
#include <sstream>
#include <string>
#include <fstream>
#include <assert.h>
#include <vector>


#include <Eigen/Dense>
#include <Eigen/Core>
#include <Eigen/SVD>

#include "model1D.h"

using Eigen::Ref;

namespace common_functions
{

// Project specific functions
void makeSyntheticModel(model1D::Model1D some_model, bool observed);
void makeWeights(model1D::Model1D some_model);
void makeMisfit(model1D::Model1D some_model);
double misfitFunctionEvaluation(model1D::Model1D some_model, const Ref<Eigen::VectorXd>& current_parameter_values, int Np, int Nr, int Ng);
double resistivityFunctionEvaluation(model1D::Model1D some_model,const Ref<Eigen::VectorXd>& parameters, std::string file_location, int Np, int Nr, int Ng);
void updateMinBetaValue(model1D::Model1D some_model, Ref<Eigen::VectorXd> min_parameter_values, const Ref<Eigen::VectorXd> parameter_values);
void makeRandomSolution(model1D::Model1D some_model, Ref<Eigen::VectorXd> solution, const Ref<Eigen::VectorXd>& min_parameter_values, const Ref<Eigen::VectorXd> max_parameter_values, const Ref<Eigen::VectorXd>& initial_category_values, const Ref<Eigen::VectorXd>& category_standard_deviation, int initial_distribution, int Np, int Nr, int Ng);
// General useful functions
void test_function(model1D::Model1D some_model); 
bool doubleArrayToFile(const double* array_in,size_t length, const std::string& file_path);
bool doubleArrayFromFile(double* array_out, size_t length, const std::string& file_path);
bool readEigenVectorFromBinaryFile(Ref<Eigen::VectorXd> vector, const std::string& filepath);
bool writeEigenVectorToBinaryFile(Ref<Eigen::VectorXd> vector, const std::string& filepath);
bool writeEigenVectorToTextFile(const Ref<Eigen::VectorXd>& vector, const std::string& filepath);
bool readEigenVectorFromTextFile(Ref<Eigen::VectorXd> vector, const std::string& filepath);
bool writeElcardinalMfil(model1D::Model1D some_model, const Ref<Eigen::VectorXd>& z_values, const Ref<Eigen::VectorXd>& res_values, std::string file_location);
int nearestIndex(const Ref<Eigen::VectorXd>& vector, double value);

template <typename Derived>
void eigenToArrayDouble(double* array_out, size_t length, const Eigen::MatrixBase<Derived> &eigen_matrix_in)
{
    array_out = eigen_matrix_in.data();    // Note: stores by column
}

template <typename Derived>
void arrayToEigenDouble(const double* array_in, size_t length, Eigen::MatrixBase<Derived> &eigen_matrix_out,int rows, int cols)
{
    if (length != rows*cols)
    {   
        std::cout << "Dimension mismatch" << std::endl;
        assert (length != rows*cols);
    }
    int i;
    int j;
    for (j = 0; j < cols; j++)
    {
        for (i = 0; i < rows; i++)
        {
            eigen_matrix_out(i,j) = array_in[j*rows + i];
        }
    }
}

template <typename Derived>
void eigenToVectorDouble(const Eigen::MatrixBase<Derived> &eigen_in, std::vector<double> &vector_out)
{
    if (vector_out.empty())
    {
        for (int i = 0; i < eigen_in.rows(); i++)
            vector_out.push_back(eigen_in(i,0));
    }
    else if (vector_out.size() == eigen_in.rows())
    {
        for (int i = 0; i < eigen_in.rows(); i++)
            vector_out[i] = eigen_in(i,0);
    }
    else 
    {   
        std::cout << "Something wrong with the dimensions in eigenToVectorDouble " << std::endl;
        return;
    }
}

template <typename Derived>
void vectorToEigenDouble(std::vector<double> vector_in, Eigen::MatrixBase<Derived> &eigen_out)
{
    for (int i = 0; i < vector_in.size(); i++)
        eigen_out(i,0) = vector_in[i];
}


bool vectorFromFileDouble(std::vector<double> &vec_out, const std::string& filepath);
bool vectorToFileDouble(const std::vector<double> &vec_in, const std::string& filepath);
void splitDouble(double d, int presision, int &integer_part, int &fractional_part);
void printDoubleVector(const std::vector<double> &vec);
void pushBackDoubleEigenMatrix(Ref<Eigen::MatrixXd> mat, const Ref<Eigen::VectorXd>& vec);
void pushBackDoubleEigenVector(Ref<Eigen::VectorXd> v, double val);

// Math related functions

double sumSquaresDouble(const std::vector<double> &v);
double pdf_gaussian(double x, double mu, double var);
double linear_interpolation(double f_a, double f_b, double a, double b, double x);
double logarithm_sum(const Ref<Eigen::VectorXd>& vector);
void dynamic_spacing(Ref<Eigen::VectorXd> z_vector, double z_min, double dz_0, double z_max, int N);
void multivariateGaussian(Ref<Eigen::VectorXd> x, const Ref<Eigen::VectorXd>& mu, const Ref<Eigen::MatrixXd>& covar);
double generateGaussianNoise(const double& variance);
double normalDistributionSample(double mean, double standard_deviation);
double correlation(const Ref<Eigen::VectorXd>& x, const Ref<Eigen::VectorXd>& y);
double sgn_double(const double& val);
double sample_variance(const Ref<Eigen::VectorXd>& x);
double getRealTime();
int noDataPoints(model1D::Model1D some_model);
double unifRand(double min, double max, bool log_uniform);
int intUnifRand(int min, int max);
void copyAndInsert(Ref<Eigen::VectorXd> new_vector, const Ref<Eigen::VectorXd>& old_vector, int index, double value);
void copyAndDelete(Ref<Eigen::VectorXd> new_vector, const Ref<Eigen::VectorXd>& old_vector, int index);
int categoryIndex(int index, int Np, int Nr, int Ng);
void inverseModelIndex(int model_index, int& np, int& nr, int& ng, int max_Np, int max_Nr, int max_Ng); 
double minPolynomialValue(model1D::Model1D some_model, int order);
double maxPolynomialValue(model1D::Model1D some_model, int order);
double meanOfBestPartOfVector(const Ref<Eigen::VectorXd>& some_vector, double percent);
void sortEigenVector(Ref<Eigen::VectorXd> some_vector);
void cumulativeFitness(const Ref<Eigen::VectorXd>& vector, Ref<Eigen::VectorXd> accumulated_vector);
void accumulate(const Ref<Eigen::VectorXd>& vector, Ref<Eigen::VectorXd> accumulated_vector);

}



#endif
