// Input to forward 1D modeler elcardinal

#include "model1D.h"
#include <string>
#include <sstream>
#include <iostream>

#include "common_functions.h"

#include <Eigen/Dense>
#include <Eigen/Core>

using Eigen::Ref;

namespace model1D
{

// Constructors:
Model1D::Model1D()
{
   set_default();
}

Model1D::Model1D(std::string name_, std::string result_type_, std::string solution_method_, double water_depth_, double water_resistivity_, std::string par_, double freq_[], int no_frequencies_, double mhor_, double zs_, double zr_, double r0_, double dr_, double rl_, std::string sortyp_, double camp_, double alen_, double azet_, double tet_, std::string method_, int acur_, std::string comp_, std::string gath_, std::string tvi_, double psi_, double x0_, double y0_, double zlay_obs_[], int zlay_obs_size_, double mver_obs_[], int mver_obs_size_, bool noise_, int run_id_, double min_misfit_value_, int no_poly_terms_, int no_resistor_terms_, int no_wavelet_terms_, double log_sum_)
{
    set_model(par_, freq_, no_frequencies, mhor_, zs_, zr_, r0_, dr_, rl_, sortyp_, camp_, alen_, azet_, tet_, method_, acur_, comp_, gath_, tvi_, psi_, x0_, y0_, zlay_obs_, zlay_obs_size_, mver_obs_, mver_obs_size_);
    set_name(name_);
    set_noise(noise_);
    set_run_id(run_id_);
    set_no_data(r0_, dr_, rl_, no_frequencies_);
    set_no_poly_terms(no_poly_terms_);
    set_no_resistor_terms(no_resistor_terms_);
    set_no_wavelet_terms(no_wavelet_terms_);
    set_no_parameters(no_poly_terms_, no_resistor_terms_, no_wavelet_terms_);
    set_log_sum(log_sum_);
    set_filenames(name_,r0_,dr_,rl_,noise_,run_id_, no_poly_terms_, no_resistor_terms_, no_wavelet_terms_, log_sum_);
    set_directories(name_, result_type_, solution_method_);
    set_min_misfit_value(min_misfit_value_);
}

// Ordinary functions:
std::string Model1D::DoubleArrayToString(double array[],int size)
{
    std::stringstream ss;
    int i=0;
    ss << array[i];
    i=1;
    while (i < size)
    {
        ss << "," << array[i];
        i++;
    }
    std::string str = ss.str();
    return str;
}

std::string Model1D::DoubleToString(double value)
{
    std::stringstream ss;
    ss << value;
    std::string str = ss.str();

    return str;
}

std::string Model1D::IntToString(int value)
{
    std::stringstream ss;
    ss << value;
    std::string str = ss.str();

    return str;
}

// Accessors and mutators:
void Model1D::set_default()
{
    name                   = "troll_001";
    intow                  = false;

    par                    = "res";
    freq                   = "0.25,0.75,1.25";  // synthetic_002,M002,M001: 0.25,1.00,2.00, else: 0.25,0.75,1.25
    mhor                   = "0.0";
    zs                     = "333.6";
    zr                     = "303.6";
    r0                     = "1000.0";
    dr                     = "100.0";
    rl                     = "10000.0";
    sortyp                 = "el";
    camp                   = "1.0";
    alen                   = "1.0";
    azet                   = "0.0";
    tet                    = "0.0";         // Towline angle (degrees) 180 is south direction
    method                 = "quad";
    acur                   = "6";
    comp                   = "ex";
    gath                   = "rec";
    tvi                    = "on";
    psi                    = "90.0";
    x0                     = "0.0";
    y0                     = "0.0";
    zlay_obs               = "1000.0,2000.0,2100.0";
    mver_obs               = "0.3125,1.0,100.0,1.0";
    
    min_misfit_value       = 1000.0;
    water_depth            = 325.6;     // based on z_r from observed data file
    water_resistivity      = 0.271;      // based on well log

    noise                  = false;
    alpha                  = 0.02;
    eta_squared            = 1e-32;
    
    run_id                 = 1000000;
    
    no_sources             = 91;
    no_receivers           = 1;
    no_frequencies         = 3;
    no_field_components    = 1;
    no_data                = 273;       // 91*1*3*1 (sources)*(recievers)*(frequencies)*(fields)
    
    no_poly_terms          = 2;         // that is, polynomial order = no_poly_terms - 1
    no_resistor_terms      = 1;         // 
    no_wavelet_terms       = 1;
    no_parameters          = 7;        // = no_poly_terms + 2*no_resistor_terms + 3*no_wavelet_terms
    log_sum                = 7.6962;     // log(100 + 2000 + 100)
    
    resistor_thickness     = 100.0;       // Fixed target thickness
    z_water_min            = 0.0;
    z_formation_max        = 2500.0;
    initial_z_spacing      = 5.0;          // Distance in meters between resistor profile evaluation points.
    z_formation_vector_length = 150;
    min_poly_contribution  = 1e-8;
    max_poly_contribution  = 100;

    initial_data_filename       = "initial_values_2_1_2_7_6962_1.txt";
    observed_field_filename     = "M001_ExObs_1000_100_10000_silent.nc";
    synthetic_field_filename    = "tmp_ExSynth_M001_silent_1000000.nc";
    weights_filename            = "M001_w_1000_100_10000_silent.nc";
    error_filename              = "tmp_ErrorFile_M001_silent_1000000.txt";
    mfil_filename               = "mfil_M001_1000000.txt";
    mfil_obs_filename           = "mfil_M001_obs.txt";
    parameter_filename          = "parameters_M001_1000000";
    min_parameter_values_filename = "min_value_parameters_M001.txt";
    max_parameter_values_filename = "max_value_parameters_M001.txt";
    solution_filename           = "solution_M001_silent_7_6962_1000000.txt";
    variable_filename           = "variables_M001_silent_7_6962_1000000.txt";

    main_dir                = "/panfs/pandora/projects/3DGN/tests/summerproject2013/master";
    bin_dir                 = "/panfs/pandora/projects/3DGN/tests/summerproject2013/master/bin";
    master_bin_dir          = "/panfs/pandora/projects/3DGN/tests/summerproject2013/master/bin";
    temporary_dump_dir      = "/scratch/oskrede/tempData";
    model_data_dir          = "/panfs/pandora/projects/3DGN/tests/summerproject2013/master/models/M001/datafiles";
    method_data_dir         = "/panfs/pandora/projects/3DGN/tests/summerproject2013/master/models/M001/simulated_annealing/data";
    initial_data_dir        = "/panfs/pandora/projects/3DGN/tests/summerproject2013/master/models/M001/datafiles/initial_values";
    mfil_temp_directory     = temporary_dump_dir;
    mfil_directory          = method_data_dir + "/profiles";
    parameter_directory     = temporary_dump_dir;
    min_parameter_values_directory = model_data_dir;
    max_parameter_values_directory = model_data_dir;
    siman_solution_dir      = method_data_dir + "/solutions";
    siman_variable_dir      = method_data_dir + "/variables";
}


void Model1D::set_model(std::string par_, double freq_[],int no_frequencies_, double mhor_, double zs_, double zr_, double r0_, double dr_, double rl_, std::string sortyp_, double camp_, double alen_, double azet_, double tet_, std::string method_, int acur_, std::string comp_, std::string gath_, std::string tvi_, double psi_, double x0_, double y0_, double zlay_obs_[], int zlay_obs_size_, double mver_obs_[], int mver_obs_size_)
{
    par                    = par_;
    freq                   = DoubleArrayToString(freq_,no_frequencies_);
    mhor                   = DoubleToString(mhor_);
    zs                     = DoubleToString(zs_);
    zr                     = DoubleToString(zr_);
    r0                     = DoubleToString(r0_);
    dr                     = DoubleToString(dr_);
    rl                     = DoubleToString(rl_);
    sortyp                 = sortyp_;
    camp                   = DoubleToString(camp_);
    alen                   = DoubleToString(alen_);
    azet                   = DoubleToString(azet_);
    tet                    = DoubleToString(tet_);
    method                 = method_;
    acur                   = IntToString(acur_);
    comp                   = comp_;
    gath                   = gath_;
    tvi                    = tvi_;
    psi                    = DoubleToString(psi_);
    x0                     = DoubleToString(x0_);
    y0                     = DoubleToString(y0_);
    zlay_obs               = DoubleArrayToString(zlay_obs_,zlay_obs_size_);
    mver_obs               = DoubleArrayToString(mver_obs_,mver_obs_size_);
}
/*
void Model1D::set_model_variables(std::string name_)
{
    name = name_;
    if (name == "troll_011")                           // troll_001 | 011     | 012    || troll_002   | 021      | 022
    {               
        zs = 325.59;                                   // 325.59    |  "      |  "     || 333.59      |          |
        zr = 295.6;                                    // 295.6     |  "      |  "     || 303.6       |          |
        x0 = 524634.6;                                 // 524634.6  |  "      |  "     || 523237.3    |          |
        y0 = 6734338.9;                                // 6734338.9 |  "      |        || 6743107.7   |          |
        intow = true;                                  //           | true    | false  ||             | true     | false
        r0 = -10000;                                   //           | -10000  | 1000   ||             | -10000   | 1000                          
        dr = 100;                                      //           | 100     | 100    ||             | 100      | 100
        rl = -1000;                                    //           | -100    | 10000  ||             | -1000    | 10000
        psi = 180.0;                                   // 180.0 (south)  | "  |  "     || 0.0 (north) |          |
        tet = 0.0;                                     // 0.0       |  "      |  "     || 180.0       |          |
        water_depth = 325.6;                           // 325.6     |  "      |  "     || 333.6       |          |
        z_formation_max = 2500.0;                      // 2500.0    |  "      |  "     || 2500.0      |          |
        water_resistivity = 0.271;                     // 0.271     |  "      |  "     || 0.271       |          |
    }
    else if (name == "troll_012")
    {
        zs = 325.59;                                   // 325.59    |  "      |  "     || 333.59      |          |
        zr = 295.6;                                    // 295.6     |  "      |  "     || 303.6       |          |
        x0 = 524634.6;                                 // 524634.6  |  "      |  "     || 523237.3    |          |
        y0 = 6734338.9;                                // 6734338.9 |  "      |        || 6743107.7   |          |
        intow = false;                                 //           | true    | false  ||             | true     | false
        r0 = 1000;                                     //           | -10000  | 1000   ||             | -10000   | 1000                          
        dr = 100;                                      //           | 100     | 100    ||             | 100      | 100
        rl = 10000;                                    //           | -100    | 10000  ||             | -1000    | 10000
        psi = 180.0;                                   // 180.0 (south)  | "  |  "     || 0.0 (north) |          |
        tet = 0.0;                                     // 0.0       |  "      |  "     || 180.0       |          |
        water_depth = 325.6;                           // 325.6     |  "      |  "     || 333.6       |          |
        z_formation_max = 2500.0;                      // 2500.0    |  "      |  "     || 2500.0      |          |
        water_resistivity = 0.271;                     // 0.271     |  "      |  "     || 0.271       |          |
    }
    else if (name == "troll_021")
    {
        zs = 333.59;                                   // 325.59    |  "      |  "     || 333.59      |          |
        zr = 303.6;                                    // 295.6     |  "      |  "     || 303.6       |          |
        x0 = 523237.3;                                 // 524634.6  |  "      |  "     || 523237.3    |          |
        y0 = 6743107.7;                                // 6734338.9 |  "      |        || 6743107.7   |          |
        intow = true;                                  //           | true    | false  ||             | true     | false
        r0 = -10000;                                   //           | -10000  | 1000   ||             | -10000   | 1000                          
        dr = 100;                                      //           | 100     | 100    ||             | 100      | 100
        rl = -1000;                                    //           | -100    | 10000  ||             | -1000    | 10000
        psi = 0.0;                                     // 180.0 (south)  | "  |  "     || 0.0 (north) |          |
        tet = 180.0;                                   // 0.0       |  "      |  "     || 180.0       |          |
        water_depth = 333.6;                           // 325.6     |  "      |  "     || 333.6       |          |
        z_formation_max = 2500.0;                      // 2500.0    |  "      |  "     || 2500.0      |          |
        water_resistivity = 0.271;                     // 0.271     |  "      |  "     || 0.271       |          |
    }
    else if (name == "troll_022")
    {
        zs = 333.59;                                   // 325.59    |  "      |  "     || 333.59      |          |
        zr = 303.6;                                    // 295.6     |  "      |  "     || 303.6       |          |
        x0 = 523237.3;                                 // 524634.6  |  "      |  "     || 523237.3    |          |
        y0 = 6743107.7;                                // 6734338.9 |  "      |        || 6743107.7   |          |
        intow = false;                                 //           | true    | false  ||             | true     | false
        r0 = 1000;                                     //           | -10000  | 1000   ||             | -10000   | 1000                          
        dr = 100;                                      //           | 100     | 100    ||             | 100      | 100
        rl = 10000;                                    //           | -100    | 10000  ||             | -1000    | 10000
        psi = 0.0;                                     // 180.0 (south)  | "  |  "     || 0.0 (north) |          |
        tet = 180.0;                                   // 0.0       |  "      |  "     || 180.0       |          |
        water_depth = 333.6;                           // 325.6     |  "      |  "     || 333.6       |          |
        z_formation_max = 2500.0;                      // 2500.0    |  "      |  "     || 2500.0      |          |
        water_resistivity = 0.271;                     // 0.271     |  "      |  "     || 0.271       |          |
    
    }
}
*/

void Model1D::set_intow(bool intow_) {intow = intow_;}

void Model1D::set_zs(double zs_) {zs = DoubleToString(zs_);}
void Model1D::set_zr(double zr_) {zr = DoubleToString(zr_);}
void Model1D::set_x0(double x0_) {x0 = DoubleToString(x0_);}
void Model1D::set_y0(double y0_) {y0 = DoubleToString(y0_);}
void Model1D::set_psi(double psi_) {psi = DoubleToString(psi_);}
void Model1D::set_tet(double tet_) {tet = DoubleToString(tet_);}
void Model1D::set_r0(double r0_) {r0 = DoubleToString(r0_);}
void Model1D::set_dr(double dr_) {dr = DoubleToString(dr_);}
void Model1D::set_rl(double rl_) {rl = DoubleToString(rl_);}

void Model1D::set_water_depth(double water_depth_) {water_depth = water_depth_;}
void Model1D::set_water_resistivity(double water_resistivity_) {water_resistivity = water_resistivity_;}
void Model1D::set_no_data(double r0_, double dr_, double rl_, double no_frequencies_)
{
    no_receivers        = 1;
    no_sources          = (int) ( (rl_ - r0_) / (dr_) + 1 );
    no_frequencies      = no_frequencies_;
    no_field_components = 1;
    no_data = no_receivers*no_sources*no_frequencies*no_field_components;
}
void Model1D::set_noise(bool noise_) {noise = noise_;}
void Model1D::set_alpha(double alpha_) {alpha = alpha_;}
void Model1D::set_eta_squared(double eta_squared_) {eta_squared = eta_squared_;}
void Model1D::set_name(std::string name_) {name = name_;}
void Model1D::set_run_id(int run_id_) {run_id = run_id_;}
void Model1D::set_no_poly_terms(int no_poly_terms_) {no_poly_terms = no_poly_terms_;}
void Model1D::set_no_resistor_terms(int no_resistor_terms_) {no_resistor_terms = no_resistor_terms_;}
void Model1D::set_no_wavelet_terms(int no_wavelet_terms_) {no_wavelet_terms = no_wavelet_terms_;}
void Model1D::set_no_parameters(int no_poly_terms_, int no_resistor_terms_, int no_wavelet_terms_)
{no_parameters = no_poly_terms_ + 2*no_resistor_terms_ + 3*no_wavelet_terms_; }
void Model1D::set_log_sum(double log_sum_) {log_sum = log_sum_;}
void Model1D::set_min_misfit_value(double min_misfit_value_) {min_misfit_value = min_misfit_value_;}

void Model1D::set_resistor_thickness(double resistor_thickness_) {resistor_thickness = resistor_thickness_;}
void Model1D::set_z_water_min(double z_water_min_) {z_water_min = z_water_min_;}
void Model1D::set_z_formation_max(double z_formation_max_) {z_formation_max = z_formation_max_;}
void Model1D::set_initial_z_spacing(double initial_z_spacing_) {initial_z_spacing = initial_z_spacing_;}
void Model1D::set_z_formation_vector_length(double z_formation_vector_length_) {z_formation_vector_length = z_formation_vector_length_;}
void Model1D::set_min_poly_contribution(double min_poly_contribution_) {min_poly_contribution = min_poly_contribution_;}
void Model1D::set_max_poly_contribution(double max_poly_contribution_) {max_poly_contribution = max_poly_contribution_;}

void Model1D::set_model_data_dir(std::string model_data_dir_) {model_data_dir = model_data_dir_;}
void Model1D::set_mfil_obs_filename(std::string mfil_obs_filename_) {mfil_obs_filename = mfil_obs_filename_;}
void Model1D::set_observed_field_filename(std::string observed_field_filename_) {observed_field_filename = observed_field_filename_;}
void Model1D::set_weights_filename(std::string weights_filename_) {weights_filename = weights_filename_;}

void Model1D::set_filenames(std::string name_, double r0_, double dr_, double rl_, bool noise_, int run_id_, int no_poly_terms_, int no_resistor_terms_, int no_wavelet_terms_, double log_sum_)
{
    std::string noise_state = (noise_ ? "noise" : "silent");
    
    std::string towing = (intow ? "intow" : "outtow");

    std::stringstream ss1;
    ss1 << run_id_;
    std::string string_run_id = ss1.str();

    std::stringstream ss2;
    ss2 << (no_poly_terms_ + 2*no_resistor_terms_ + 3*no_wavelet_terms_);
    std::string string_no_parameters = ss2.str();

    int integer_part;
    int fractional_part;
    common_functions::splitDouble(log_sum_, 4, integer_part, fractional_part);

    std::string s_ip                = IntToString(integer_part);
    std::string s_fp                = IntToString(fractional_part);
    std::string r0__                = IntToString(static_cast<int>(r0_));
    std::string dr__                = IntToString(static_cast<int>(dr_));
    std::string rl__                = IntToString(static_cast<int>(rl_));
    std::string no_poly_terms       = IntToString(no_poly_terms_);
    std::string no_resistor_terms   = IntToString(no_resistor_terms_);
    std::string no_wavelet_terms    = IntToString(no_wavelet_terms_);
    
    if (name_ == "troll_001")
    {
        initial_data_filename       = "initial_values_" + name + "_" + s_ip + "_" + s_fp + "_" + string_run_id + ".txt";
        observed_field_filename     = "ST0755_02Tx003a_02Rx020a_data_1000_100_10000_025_075_125_Ex.nc";
        original_observed_field_filename = "ST0755_02Tx003a_02Rx020a_data.nc";
        synthetic_field_filename    = "tmp_ExSynth_ST0755_02Tx003a_02Rx020a_data_" + string_run_id + ".nc";
        temp_weights_filename       = "tmp_weights.nc";
        weights_filename            = "weights_ST0755_02Tx003a_02Rx020a_data_1000_100_10000_025_075_125_1500_1500_1000_9000_8000_6000.nc";
        error_filename              = "tmp_ErrorFile_" + name_ + "_" + noise_state + "_" + string_run_id + ".txt";
        mfil_filename               = "mfil_" + name_ + "_" + string_run_id + ".txt";
        mfil_obs_filename           = "31_2-5.las";
    }
    else if (name_ == "troll_002")
    {
        initial_data_filename       = "initial_values_" + name + "_" + s_ip + "_" + s_fp + "_" + string_run_id + ".txt";
        observed_field_filename     = "ST0755_02Tx010a_02Rx018a_data_1000_100_10000_025_075_125_Ex.nc";
        original_observed_field_filename = "ST0755_02Tx010a_02Rx018a_data.nc";
        synthetic_field_filename    = "tmp_ExSynth_ST0755_02Tx010a_02Rx018a_data_" + string_run_id + ".nc";
        weights_filename            = "weights_ST0755_02Tx010a_02Rx018a_data_1000_100_10000_025_075_125_1500_1500_1000_9000_8000_6000.nc";
        temp_weights_filename       = "tmp_weights.nc";
        error_filename              = "tmp_ErrorFile_" + name_ + "_" + noise_state + "_" + string_run_id + ".txt";
        mfil_filename               = "mfil_" + name_ + "_" + string_run_id + ".txt";
        mfil_obs_filename           = "31_2-5.las";
    }
    else if (name_ == "troll_011" || name_ == "troll_012")
    {
        initial_data_filename       = "initial_values_" + name + "_" + s_ip + "_" + s_fp + "_" + string_run_id + ".txt";
        observed_field_filename     = "ST0755_02Tx003a_02Rx020a_data_1000_100_10000_025_075_125_Ex_" + towing + ".nc";
        original_observed_field_filename = "ST0755_02Tx003a_02Rx020a_data.nc";
        synthetic_field_filename    = "tmp_ExSynth_ST0755_02Tx003a_02Rx020a_data_" + string_run_id + ".nc";
        weights_filename            = "weights_ST0755_02Tx003a_02Rx020a_data_1000_100_10000_025_075_125_1500_1500_1000_9000_8000_6000_" + towing + ".nc";
        temp_weights_filename       = "tmp_weights.nc";
        error_filename              = "tmp_ErrorFile_" + name_ + "_" + towing + "_" + noise_state + "_" + string_run_id + ".txt";
        mfil_filename               = "mfil_" + name_ + "_" + towing + "_" + string_run_id + ".txt";
        mfil_obs_filename           = "31_2-5.las";
    }
    else if (name_ == "troll_021" || name_ == "troll_022")
    {
        initial_data_filename       = "initial_values_" + name + "_" + s_ip + "_" + s_fp + "_" + string_run_id + ".txt";
        observed_field_filename     = "ST0755_02Tx010a_02Rx018a_data_1000_100_10000_025_075_125_Ex_" + towing + ".nc";
        original_observed_field_filename = "ST0755_02Tx010a_02Rx018a_data.nc";
        synthetic_field_filename    = "tmp_ExSynth_ST0755_02Tx010a_02Rx018a_data_" + string_run_id + ".nc";
        weights_filename            = "weights_ST0755_02Tx010a_02Rx018a_data_1000_100_10000_025_075_125_1500_1500_1000_9000_8000_6000_" + towing + ".nc";
        temp_weights_filename       = "tmp_weights.nc";
        error_filename              = "tmp_ErrorFile_" + name_ + "_" + towing + "_" + noise_state + "_" + string_run_id + ".txt";
        mfil_filename               = "mfil_" + name_ + "_" + towing + "_" + string_run_id + ".txt";
        mfil_obs_filename           = "31_2-5.las";
    }
    else
    {
        initial_data_filename       = "initial_values_" + name + "_" + s_ip + "_" + s_fp + "_" + string_run_id + ".txt";
        observed_field_filename     = name_ + "_ExObs_" + r0__ + "_" + dr__ + "_" + rl__ + "_" + noise_state + ".nc";
        original_observed_field_filename = "";
        synthetic_field_filename    = "tmp_ExSynth_" + name_ + "_" + noise_state + "_" + string_run_id + ".nc";
        weights_filename            = name_ + "_w_" + r0__ + "_" + dr__ + "_" + rl__ + "_" + noise_state + ".nc";
        temp_weights_filename       = "tmp_weights.nc";
        error_filename              = "tmp_ErrorFile_" + name_ + "_" + noise_state + "_" + string_run_id + ".txt";
        mfil_filename               = "mfil_" + name_ + "_" + string_run_id + ".txt";
        mfil_obs_filename           = "mfil_obs" + name_ + ".txt";
    }
    
    //mfil_filename = "mfil_M001_1220181997.txt";
    parameter_filename          = "parameters_" + name_ + "_" + string_run_id;
    min_parameter_values_filename = "min_value_parameters_" + name_ + ".txt";
    max_parameter_values_filename = "max_value_parameters_" + name_ + ".txt";
    solution_filename           = "solution_" + name_ + "_" + noise_state + "_" + s_ip + "_" + s_fp + "_" + string_run_id + ".txt";
    variable_filename           = "variables_" + name_ + "_" + noise_state + "_" + s_ip + "_" + s_fp + "_" + string_run_id + ".txt";
}

void Model1D::set_directories(std::string name_, std::string result_type_, std::string solution_method_)
{
    //name_ is either troll_011,troll_012,troll_021,troll_022
    //result_type is either tests or final_results
    //solution_method is either fixed, dyndim or transdim
    //

    main_dir                = "/panfs/pandora/projects/3DGN/tests/summerproject2013/master";
    bin_dir                 = "/panfs/pandora/projects/3DGN/tests/summerproject2013/master/bin";
    master_bin_dir          = "/panfs/pandora/projects/3DGN/tests/summerproject2013/master/bin";
    temporary_dump_dir      = "/scratch/oskrede/tempData";
    model_data_dir          = "/panfs/pandora/projects/3DGN/tests/summerproject2013/master/models/" + name_ + "/datafiles";
    method_data_dir         = "/panfs/pandora/projects/3DGN/tests/summerproject2013/master/models/" + name_ + "/simulated_annealing/data";
    initial_data_dir        = "/panfs/pandora/projects/3DGN/tests/summerproject2013/master/models/" + name_ + "/datafiles/initial_values";
    parameter_directory     = temporary_dump_dir;
    min_parameter_values_directory = model_data_dir;
    max_parameter_values_directory = model_data_dir;
    mfil_temp_directory     = temporary_dump_dir;
    mfil_directory          = method_data_dir + "/" + result_type_ + "/" + solution_method_ + "/profiles";
    siman_solution_dir      = method_data_dir + "/" + result_type_ + "/" + solution_method_ + "/solutions";
    siman_variable_dir      = method_data_dir + "/" + result_type_ + "/" + solution_method_ + "/variables";
}

void Model1D::new_run(std::string name_, std::string result_type_, std::string solution_method_, bool noise_, int run_id_, int no_poly_terms_, int no_resistor_terms_, int no_wavelet_terms_, double log_sum_)
{
    // update run_id and filenames for a new run according to the argument
    // run_id_, no_poly_terms, no_resistor_terms, no_wavelet_terms and log_sum_,  that is, everything else is held constant
    
    double d_r0;
    double d_dr;
    double d_rl;
    std::stringstream convert_r0(r0);
    std::stringstream convert_dr(dr);
    std::stringstream convert_rl(rl);
    if ( !(convert_r0 >> d_r0) )
        d_r0 = 0.0;
    if ( !(convert_dr >> d_dr) )
        d_dr = 0.0;
    if ( !(convert_rl >> d_rl) )
        d_rl = 0.0;
  
    set_name(name_);
    set_noise(noise_);
    set_no_poly_terms(no_poly_terms_);
    set_no_resistor_terms(no_resistor_terms_);
    set_no_wavelet_terms(no_wavelet_terms_);
    set_no_parameters(no_poly_terms_, no_resistor_terms_, no_wavelet_terms_);
    set_log_sum(log_sum_);
    set_run_id(run_id_);
    set_filenames(name_, d_r0, d_dr, d_rl, noise, run_id, no_poly_terms_, no_resistor_terms, no_wavelet_terms_, log_sum_);
    set_directories(name_,result_type_,solution_method_);
}

}
