#include <iostream>     // std::cout,std::cin,std::endl;
#include <string>       // std::string
#include <sstream>      // std::stringstream
#include <vector>       // std::vector<>
#include <time.h>       // time_t, struct tm, difftime, time, mktime 
#include <ctime>        // 
#include <cstdlib>      // RAND_MAX
#include <stdlib.h>     // srand, rand
#include <math.h>       // sqrt, exp, pow, log
#include <numeric>      // std::accumulate,
#include <algorithm>    // std::transform, std::max
#include <iomanip>      // std::setw,

#include <Eigen/Dense>
#include <Eigen/Core>
#include <Eigen/SVD>

#include <boost/lambda/lambda.hpp>
#include <boost/math/distributions/chi_squared.hpp>

#include "model1D.h"
#include "common_functions.h"
#include "simulated_annealing.h"

using boost::math::chi_squared;
using boost::math::quantile;
using std::cout;
using std::endl;
using Eigen::Ref;

int main(int argc, char *argv[])
{
    double start_time, end_time;
    start_time = common_functions::getRealTime();

    // Command line arguments:
    std::string initial_filepath = argv[1];
    int no_parameters = atoi (argv[2]);
    int run_id = atoi (argv[3]);
    std::stringstream ss1;
    ss1 << run_id;
    std::string string_run_id = ss1.str();

    std::string name;
    bool noise;
    int no_poly_terms;
    int no_resistor_terms;
    int no_wavelet_terms;
    double zs;
    double zr;
    double x0;
    double y0;
    bool intow;
    double r0;                          
    double dr;
    double rl;
    double psi;
    double tet;
    double water_depth;
    double z_formation_max;
    double water_resistivity;
    Eigen::VectorXd parameters(no_parameters);

    std::ifstream initial_in_stream(initial_filepath.c_str(), std::ios::in);
    if (!initial_in_stream.is_open())
        std::cout << "Unable to open file()" << endl;
    initial_in_stream >> name;
    initial_in_stream >> noise;
    initial_in_stream >> no_poly_terms;
    initial_in_stream >> no_resistor_terms;
    initial_in_stream >> no_wavelet_terms;
    initial_in_stream >> zs;
    initial_in_stream >> zr;
    initial_in_stream >> x0;
    initial_in_stream >> y0;
    initial_in_stream >> intow;
    initial_in_stream >> r0;
    initial_in_stream >> dr;
    initial_in_stream >> rl;
    initial_in_stream >> psi;
    initial_in_stream >> tet;
    initial_in_stream >> water_depth;
    initial_in_stream >> z_formation_max;
    initial_in_stream >> water_resistivity;
        
    for (int i = 0; i < no_parameters; ++i)
        initial_in_stream >> parameters[i];
    initial_in_stream.close();
    model1D::Model1D run_model; // invoke the default constructor, apply changes with set functions etc.

    run_model.set_zs(zs);
    run_model.set_zr(zr);
    run_model.set_x0(x0);
    run_model.set_y0(y0);
    run_model.set_intow(intow);
    run_model.set_r0(r0);                          
    run_model.set_dr(dr);
    run_model.set_rl(rl);
    run_model.set_psi(psi);
    run_model.set_tet(tet);
    run_model.set_water_depth(water_depth);
    run_model.set_z_formation_max(z_formation_max);
    run_model.set_water_resistivity(water_resistivity);

    if ((parameters.size() != no_parameters) && (parameters.size() != (no_poly_terms + 2*no_resistor_terms + 3*no_wavelet_terms)))
        cout << "Something is wrong with parameter no_parameters." << endl;

    double log_sum = common_functions::logarithm_sum(parameters);
    // Update information about the new run. Invokes the set functions
    // set_run_id();
    // set_no_poly_terms();
    // set_no_poly_terms();
    // set_no_wavelet_terms();
    // set_filenames();
    // se_log_sum();
    run_model.new_run(name, noise, run_id, no_poly_terms, no_resistor_terms, no_wavelet_terms, log_sum);
    //run_model.set_initial_z_spacing(10.0);
    //run_model.set_z_formation_vector_length(100);

    int no_data_points = common_functions::noDataPoints(run_model);
    
    
    simulated_annealing::SimulatedAnnealing this_run;
    this_run.set_num_iter_before_update(15*no_parameters); 

    std::vector<double> all_misfit_values;           // std::vector because of push_back
    std::vector<double> latest_misfit_values;
    double sum_latest_misfit_values;
    Eigen::VectorXd direction = Eigen::MatrixXd::Zero(parameters.size(),1);
    Eigen::VectorXd current_solution = parameters;     // Parhaps unnecessary
    Eigen::VectorXd proposed_solution = current_solution;

    // Parameter fix
    //current_solution(0) = 0.0;
    //current_solution(1) = 0.002;
    
    // Linear constraints:
    std::string min_parameter_values_filepath   = run_model.get_min_parameter_values_directory() + "/" + run_model.get_min_parameter_values_filename();
    std::string max_parameter_values_filepath   = run_model.get_max_parameter_values_directory() + "/" + run_model.get_max_parameter_values_filename();
    Eigen::VectorXd min_parameter_values(no_parameters);
    Eigen::VectorXd max_parameter_values(no_parameters);
    common_functions::readEigenVectorFromTextFile(min_parameter_values,min_parameter_values_filepath);
    common_functions::readEigenVectorFromTextFile(max_parameter_values,max_parameter_values_filepath);

    double step_length = 0.0;
    int num_halves = 0;         // Number of times the max step length sample threshold is halved
   
    //cout << current_solution << endl;

    double current_misfit_value = common_functions::misfitFunctionEvaluation(run_model, current_solution);
    double current_RMS_value = sqrt(current_misfit_value/no_data_points);
    double proposed_misfit_value = current_misfit_value;
    double proposed_RMS_value = current_RMS_value;

    all_misfit_values.push_back(current_misfit_value);
    latest_misfit_values.push_back(current_misfit_value);

    int no_iter_history = this_run.get_no_iter_history();
    Eigen::MatrixXd all_latest_solution_values = Eigen::MatrixXd::Zero(no_iter_history,no_parameters);
    Eigen::VectorXd all_latest_misfit_values = Eigen::MatrixXd::Zero(no_iter_history,1);

    common_functions::pushBackDoubleEigenMatrix(all_latest_solution_values,current_solution);
    common_functions::pushBackDoubleEigenVector(all_latest_misfit_values,current_misfit_value);
    
    double min_misfit_value = run_model.get_min_misfit_value();
    // initial temperature
    chi_squared mydist(no_parameters);  // degrees of freedom = no_parameters
    double percentile = quantile(mydist,1 - this_run.get_p_chi());
    //double initial_temperature = 2*(current_misfit_value - min_misfit_value)/percentile;
    double initial_temperature = 10000;
    double temperature = initial_temperature;

    bool update_temperature = false;
    double accept_prob = 0.0;
    double uniform_0_1;
    double latest_misfit_min;
    double latest_misfit_mean;
    double reduce_step_length_const;
    bool accept_proposed_when_smaller = false;
    int accept_proposed_when_smaller_counter = 0;
    int accept_proposed_counter = 0;
    double accept_ratio = 0;

    int global_iterator = 0;
    int local_iterator = 0;

    //cout << current_RMS_value << endl;
    
    end_time = common_functions::getRealTime();
    std::string solution_filepath = run_model.get_siman_solution_dir() + "/" + run_model.get_solution_filename();
    std::ofstream out_stream(solution_filepath.c_str(), std::ios::out | std::ios::trunc);
    if (!out_stream.is_open())
        std::cout << "Unable to open file()" << endl;
    out_stream << std::setw(7) << "Iter" << std::setw(12) << "Misfit" << std::setw(9) << "RMS" << std::setw(9) << "Prop RMS";
    out_stream << std::setw(13) << "Accept prob" << std::setw(13) << "Accept ratio" << std::setw(13) << "Temperature" << std::setw(13) << "Step length" << std::setw(13) << "Time" << std::setw(3) << "n\n";
    out_stream << std::setw(7) << global_iterator << std::setw(12) << current_misfit_value << std::setw(9) << current_RMS_value;
    out_stream << std::setw(9) << proposed_RMS_value << std::setw(13)   << accept_prob << std::setw(13) << accept_ratio << std::setw(13) << temperature << std::setw(13) << "88.88e-4" << std::setw(13) << end_time - start_time << std::setw(3) << num_halves;
   
    for (int i = 0; i < no_parameters; ++i)
        out_stream << std::setw(15) << current_solution(i);
    out_stream << "\n";

    double min_resistivity_value = run_model.get_water_resistivity();
    double current_resistivity_min = 0.0;
    bool current_resistivity_is_smaller = true;

    std::string variable_filepath = run_model.get_siman_variable_dir() + "/" + run_model.get_variable_filename();
    this_run.writeVariableValuesToTxtFile(run_model, parameters, variable_filepath);
    
    bool proceed = true;
    
    //cout << "initial_values: " << endl << initial_values << endl;
    //cout << "current_solution: " << endl << current_solution << endl;
    //cout << "proposed_solution: " << endl << proposed_solution << endl;
    //cout << "direction: " << endl << direction << endl;
    while (proceed)
    {
        while (current_resistivity_is_smaller)
        {
            this_run.newProposedSolution(run_model, proposed_solution, current_solution, min_parameter_values, max_parameter_values, all_latest_solution_values, all_latest_misfit_values, num_halves, global_iterator, no_poly_terms, no_resistor_terms, no_wavelet_terms);
            current_resistivity_min = common_functions::resistivityFunctionEvaluation(run_model, proposed_solution, run_model.get_temporary_dump_dir() + "/temp_profile_" + string_run_id); 
            current_resistivity_is_smaller = current_resistivity_min < min_resistivity_value;
            //cout << min_resistivity_value << endl;
            //cout << current_resistivity_min << endl;
            //cout << current_resistivity_is_smaller << endl;
        }
        current_resistivity_is_smaller = true;

        proposed_misfit_value = common_functions::misfitFunctionEvaluation(run_model, proposed_solution);
        
        common_functions::pushBackDoubleEigenMatrix(all_latest_solution_values,proposed_solution);
        common_functions::pushBackDoubleEigenVector(all_latest_misfit_values,proposed_misfit_value);
        if (proposed_misfit_value <= current_misfit_value)
        {
            current_solution = proposed_solution;
            current_misfit_value = proposed_misfit_value;
            //update_temperature = true;
            accept_prob = 1.0;
            accept_proposed_counter++;

        }
        else
        {
            uniform_0_1 = ((double)rand()/(double)RAND_MAX);
            accept_prob = exp(- (proposed_misfit_value - current_misfit_value)/ temperature );
            if (uniform_0_1 < accept_prob)
            {
                current_solution = proposed_solution;
                current_misfit_value = proposed_misfit_value;
                //update_temperature = true;
                accept_proposed_counter++;
            }
        }

        accept_ratio = (double)accept_proposed_counter/((double)global_iterator + 1.0);
        proposed_RMS_value = sqrt(proposed_misfit_value/no_data_points);
        current_RMS_value = sqrt(current_misfit_value/no_data_points);

        all_misfit_values.push_back(current_misfit_value);
        latest_misfit_values.push_back(current_misfit_value);

        if (update_temperature)
        {
            temperature = this_run.updateTemperature(run_model, all_misfit_values, min_misfit_value, no_parameters, temperature, initial_temperature, global_iterator);
            update_temperature = false;
        }

        if (local_iterator >= this_run.get_num_iter_before_update())
        {
            update_temperature = true;
            local_iterator = 0;
        }

        if (this_run.get_reduce_step_length() && (local_iterator > this_run.get_num_iter_before_update()))
        {
            latest_misfit_min = *std::min_element(latest_misfit_values.begin(), latest_misfit_values.end());
            sum_latest_misfit_values = std::accumulate(latest_misfit_values.begin(), latest_misfit_values.end(), 0.0);
            latest_misfit_mean = sum_latest_misfit_values/latest_misfit_values.size();
            reduce_step_length_const = (latest_misfit_mean - latest_misfit_min)/latest_misfit_mean;
            if (reduce_step_length_const < this_run.get_update_threshold())
            {
                num_halves++;
            }
            local_iterator = 0;
            latest_misfit_values.clear();
            latest_misfit_values.push_back(current_misfit_value);
            update_temperature = true;
        }

        end_time = common_functions::getRealTime();
        out_stream << std::setw(7) << global_iterator << std::setw(12) << current_misfit_value << std::setw(9) << current_RMS_value;
        out_stream << std::setw(9) << proposed_RMS_value << std::setw(13)   << accept_prob << std::setw(13) << accept_ratio << std::setw(13) << temperature << std::setw(13) << step_length << std::setw(13) << end_time - start_time << std::setw(3) << num_halves;
   
        for (int i = 0; i < no_parameters; ++i)
            out_stream << std::setw(15) << current_solution(i);
        out_stream << "\n";
        
        // Terminate?

        if (this_run.get_max_iter_termination() && (global_iterator > this_run.get_max_iter() )) // this_run.get_max_iter() 
        {
            proceed = false;
            cout << "max iteration termination" << endl;
        }

        if (this_run.get_fn_value_termination())
        {
            if (!run_model.get_noise())
            {
                if ((current_RMS_value < 1.0))  
                {   
                    proceed = false;
                    cout << "silent function value termination" << endl;
                }
            }
            else
            {
                if ((current_RMS_value < 1.0))
                {
                    proceed = false;
                    cout << "noise function value termination" << endl;
                }
            }
        }

        global_iterator++;
        local_iterator++;

        common_functions::resistivityFunctionEvaluation(run_model,current_solution,"synthetic_keep");
        
        cout << "Iter nr: " << global_iterator << "   RMS = " << current_RMS_value << endl;
        // cout << "current_solution: " << endl << current_solution << endl;
        // cout << "proposed_solution: " << endl << proposed_solution << endl;
        // cout << "direction: " << endl << direction << endl;
    }

    out_stream.close();


    end_time = common_functions::getRealTime();
    cout << "Elapsed time (s):  " << end_time - start_time << endl; 
    cout << "Solution in: " << solution_filepath << endl;
    cout << "Variable values in: " << variable_filepath << endl;

    return 0;
}
