#include "simulated_annealing.h"

#include <iostream>
#include <sstream>
#include <fstream>
#include <string>
#include <algorithm>
#include <vector>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <cstdlib>

#include <Eigen/Dense>
#include <Eigen/Core>
#include <Eigen/SVD>

#include <boost/lambda/lambda.hpp>
#include <boost/math/distributions/chi_squared.hpp>

#include "common_functions.h"
#include "model1D.h"

using boost::math::chi_squared;
using boost::math::quantile;
using Eigen::Ref;
using std::cout;
using std::endl;

namespace simulated_annealing
{

// Constructors:
SimulatedAnnealing::SimulatedAnnealing()
{    
    max_iter_termination    = true;
    max_iter                = 20000;
    fn_value_termination    = true;
    reduce_step_length      = false;
    num_iter_before_update  = 10*15;
    update_threshold        = 1e-3;
    no_iter_history         = 10;
    direction_type          = 1;
    temperature_schedule    = 3;        // 1: romeijn, 2: geometric, 3: lundy
    update_temp_while_accept= true;
    p_chi                   = 0.01;            
    geometric_temp_const    = 0.925;
    lundy_temp_const        = 5e-5;
    poly_term_log_sample    = 1;        // 1: logarithmic uniform direction sampler, 2: logarithmic uniform step length sampler
    new_solution_method     = 2;        // 1: old way, 2: new, 3: EMGS
    update_single_parameter = true;
    dynamic_dimensions      = true;        
    min_Np                  = 0;        // minimum order of polynomial term, 0 means just the constant term
    min_Nr                  = 0;        // minimum number of resistors
    min_Ng                  = 0;        // minimum number of gaussfunctions
    max_Np                  = 10;       // maximum order of polynomial term
    max_Nr                  = 10;       // maximum number of resistors
    max_Ng                  = 10;       // maximum number of resistors
    p_nothing               = 0.4;      // probability of not changing number of parameters
    p_in_Np                 = 0.1;      // probability of increase Np by one       
    p_de_Np                 = 0.1;      // probability of decrease Np by one
    p_in_Nr                 = 0.1;
    p_de_Nr                 = 0.1;
    p_in_Ng                 = 0.1;
    p_de_Ng                 = 0.1;
    no_change               = true;
    increase_Np             = false;
    decrease_Np             = false;
    increase_Nr             = false;
    decrease_Nr             = false;
    increase_Ng             = false;
    decrease_Ng             = false;
    dyndim_lifo             = true;
    
    // variables unique to the new transverse simulated annealing
    best_model_prob         = 0.8;
}

SimulatedAnnealing::SimulatedAnnealing(double p_chi_, bool max_iter_termination_, int max_iter_, bool fn_value_termination_, bool reduce_step_length_, int num_iter_before_update_, double update_threshold_)
{
    p_chi                   = p_chi_;
    max_iter_termination    = max_iter_termination_;
    max_iter                = max_iter_;
    fn_value_termination    = fn_value_termination_;
    reduce_step_length      = reduce_step_length_;
    num_iter_before_update  = num_iter_before_update_;
    update_threshold        = update_threshold_;
}    


// Regular functions:
double SimulatedAnnealing::updateTemperature(model1D::Model1D some_model, std::vector<double> misfit_values, double min_misfit_value, int no_parameters, double current_temperature, double initial_temperature, double global_iterate)
{
    double T;

    if (temperature_schedule == 1)
    {
        chi_squared mydist(no_parameters);  // degrees of freedom = no_parameters
        double percentile = quantile(mydist,1 - p_chi);
        double current_min = *std::min_element(misfit_values.begin(), misfit_values.end());
        T = 2*(current_min - min_misfit_value)/percentile;
    }
    else if (temperature_schedule == 2)
    {
        T = initial_temperature*pow(geometric_temp_const,(double) global_iterate);
    }
    else if (temperature_schedule == 3)
    {
        T = initial_temperature/(1 + global_iterate*lundy_temp_const*initial_temperature);
    }
    else if (temperature_schedule == 4)
    {
        T = 0.1*(initial_temperature*pow(geometric_temp_const,(double) global_iterate)) + 0.9*(initial_temperature/(1 + global_iterate*lundy_temp_const*initial_temperature));
    }
        
    return T;
}

void SimulatedAnnealing::newProposedSolution(model1D::Model1D some_model, Ref<Eigen::VectorXd> proposed_solution, const Ref<Eigen::VectorXd>& current_solution, const Ref<Eigen::VectorXd>& min_parameter_values, const Ref<Eigen::VectorXd>& max_parameter_values, const Ref<Eigen::MatrixXd>& latest_solution_values, const Ref<Eigen::VectorXd>& latest_misfit_values, int num_halves, int global_iterator, int& proposed_Np, int& proposed_Nr, int& proposed_Ng, int Np, int Nr, int Ng, const Ref<Eigen::VectorXd>& min_step_length_values, const Ref<Eigen::VectorXd>& max_step_length_values, const Ref<Eigen::VectorXd>& initial_category_values, const Ref<Eigen::VectorXd>& category_standard_deviation, int initial_distribution)
{
    int Np_new = Np;
    int Nr_new = Nr;
    int Ng_new = Ng;
    int i,j,k;
    int category_index;
    int no_parameters = 1 + Np + 2*Nr + 3*Ng;
    int no_parameters_new = no_parameters;
    
    rand();
    rand();
    rand();
    double rand01_1 = (double) rand() / RAND_MAX;
    double rand01_2 = (double) rand() / RAND_MAX;
    double rand01_3 = (double) rand() / RAND_MAX;
    double rand01_4 = (double) rand() / RAND_MAX;
    double rand01_5 = (double) rand() / RAND_MAX;
    double rand01_6 = (double) rand() / RAND_MAX;
    double rand01_7 = (double) rand() / RAND_MAX;
    double c_min;
    double c_max;

    if (dynamic_dimensions)
    {
        if (!no_change)
        {
            double new_value_1;
            double new_value_2;
            double new_value_3;
            int remove_index_1;
            int remove_index_2;
            int remove_index_3;
            if (increase_Np)
            {
                Eigen::VectorXd temp_vector_p1 = Eigen::MatrixXd::Zero(1 + Np + 2*Nr + 3*Ng + 1,1);
                Np_new = Np + 1;
                //c_min = common_functions::minPolynomialValue(some_model,Np_new);
                //c_max = common_functions::maxPolynomialValue(some_model,Np_new);
                //new_value_1 = common_functions::unifRand(c_min,c_max,false);
                if (initial_distribution == 1)
                {
                    new_value_1 = common_functions::unifRand(min_parameter_values(1),max_parameter_values(1),false);
                }
                else if (initial_distribution == 2)
                {
                    new_value_1 = common_functions::normalDistributionSample(initial_category_values(1),category_standard_deviation(1));
                    if (new_value_1 < min_parameter_values(1))
                        new_value_1 = min_parameter_values(1);
                    else if (new_value_1 > max_parameter_values(1))
                        new_value_1 = max_parameter_values(1);
                }
                    common_functions::copyAndInsert(temp_vector_p1,current_solution,Np_new,new_value_1);
                proposed_solution = temp_vector_p1;
            }
            else if (decrease_Np)
            {
                Eigen::VectorXd temp_vector_m1 = Eigen::MatrixXd::Zero(1 + Np + 2*Nr + 3*Ng - 1,1);
                Np_new = Np - 1;
                if (dyndim_lifo)
                {
                    common_functions::copyAndDelete(temp_vector_m1,current_solution,Np);
                }
                else
                {
                    remove_index_1 = common_functions::intUnifRand(1,Np); 
                    common_functions::copyAndDelete(temp_vector_m1,current_solution,remove_index_1);
                }
                proposed_solution = temp_vector_m1;
            }
            else if (increase_Nr)
            {
                Eigen::VectorXd temp_vector_p1 = Eigen::MatrixXd::Zero(1 + Np + 2*Nr + 3*Ng + 1,1);
                Eigen::VectorXd temp_vector_p2 = Eigen::MatrixXd::Zero(1 + Np + 2*Nr + 3*Ng + 2,1);
                Nr_new = Nr + 1;
                if (initial_distribution == 1)
                {
                    new_value_1 = common_functions::unifRand(min_parameter_values(2),max_parameter_values(2),false);
                    new_value_2 = common_functions::unifRand(min_parameter_values(3),max_parameter_values(3),false);
                }
                else if (initial_distribution == 2)
                {
                    new_value_1 = common_functions::normalDistributionSample(initial_category_values(2),category_standard_deviation(2));
                    new_value_2 = common_functions::normalDistributionSample(initial_category_values(3),category_standard_deviation(3));
                    if (new_value_1 < min_parameter_values(2))
                        new_value_1 = min_parameter_values(2);
                    else if (new_value_1 > max_parameter_values(2))
                        new_value_1 = max_parameter_values(2);
                    if (new_value_2 < min_parameter_values(3))
                        new_value_2 = min_parameter_values(3);
                    else if (new_value_2 > max_parameter_values(3))
                        new_value_2 = max_parameter_values(3);
                }
                common_functions::copyAndInsert(temp_vector_p1,current_solution,Np + Nr_new,new_value_1);
                common_functions::copyAndInsert(temp_vector_p2,temp_vector_p1,Np + 2*Nr_new,new_value_2);
                proposed_solution = temp_vector_p2;
            }
            else if (decrease_Nr)
            {
                Eigen::VectorXd temp_vector_m1 = Eigen::MatrixXd::Zero(1 + Np + 2*Nr + 3*Ng - 1,1);
                Eigen::VectorXd temp_vector_m2 = Eigen::MatrixXd::Zero(1 + Np + 2*Nr + 3*Ng - 2,1);
                Nr_new = Nr - 1;
                if (dyndim_lifo)
                {
                    common_functions::copyAndDelete(temp_vector_m1,current_solution,Np + 2*Nr);
                    common_functions::copyAndDelete(temp_vector_m2,temp_vector_m1,Np + Nr);
                }
                else
                {
                    remove_index_1 = common_functions::intUnifRand(Np + 1,Np + Nr); 
                    remove_index_2 = common_functions::intUnifRand(Np + Nr + 1,Np + 2*Nr); 
                    common_functions::copyAndDelete(temp_vector_m1,current_solution,remove_index_2);
                    common_functions::copyAndDelete(temp_vector_m1,current_solution,remove_index_1);
                }
                proposed_solution = temp_vector_m2;
            }
            else if (increase_Ng)
            {
                Eigen::VectorXd temp_vector_p1 = Eigen::MatrixXd::Zero(1 + Np + 2*Nr + 3*Ng + 1,1);
                Eigen::VectorXd temp_vector_p2 = Eigen::MatrixXd::Zero(1 + Np + 2*Nr + 3*Ng + 2,1);
                Eigen::VectorXd temp_vector_p3 = Eigen::MatrixXd::Zero(1 + Np + 2*Nr + 3*Ng + 3,1);
                Ng_new = Ng + 1;
                if (initial_distribution == 1)
                {
                    new_value_1 = common_functions::unifRand(min_parameter_values(4),max_parameter_values(4),false);
                    new_value_2 = common_functions::unifRand(min_parameter_values(5),max_parameter_values(5),false);
                    new_value_3 = common_functions::unifRand(min_parameter_values(6),max_parameter_values(6),false);
                }
                else if (initial_distribution == 2)
                {
                    new_value_1 = common_functions::normalDistributionSample(initial_category_values(4),category_standard_deviation(4));
                    new_value_2 = common_functions::normalDistributionSample(initial_category_values(5),category_standard_deviation(5));
                    new_value_3 = common_functions::normalDistributionSample(initial_category_values(6),category_standard_deviation(6));
                    if (new_value_1 < min_parameter_values(4))
                        new_value_1 = min_parameter_values(4);
                    else if (new_value_1 > max_parameter_values(4))
                        new_value_1 = max_parameter_values(4);
                    if (new_value_2 < min_parameter_values(5))
                        new_value_2 = min_parameter_values(5);
                    else if (new_value_2 > max_parameter_values(5))
                        new_value_2 = max_parameter_values(5);
                    if (new_value_3 < min_parameter_values(6))
                        new_value_3 = min_parameter_values(6);
                    else if (new_value_3 > max_parameter_values(6))
                        new_value_3 = max_parameter_values(6);
                }
                common_functions::copyAndInsert(temp_vector_p1,current_solution,Np + 2*Nr + Ng_new,new_value_1);
                common_functions::copyAndInsert(temp_vector_p2,temp_vector_p1,Np + 2*Nr + 2*Ng_new,new_value_2);
                common_functions::copyAndInsert(temp_vector_p3,temp_vector_p2,Np + 2*Nr + 3*Ng_new,new_value_3);
                proposed_solution = temp_vector_p3;
            }
            else if (decrease_Ng)
            {
                Eigen::VectorXd temp_vector_m1 = Eigen::MatrixXd::Zero(1 + Np + 2*Nr + 3*Ng - 1,1);
                Eigen::VectorXd temp_vector_m2 = Eigen::MatrixXd::Zero(1 + Np + 2*Nr + 3*Ng - 2,1);
                Eigen::VectorXd temp_vector_m3 = Eigen::MatrixXd::Zero(1 + Np + 2*Nr + 3*Ng - 3,1);
                Ng_new = Ng - 1;
                common_functions::copyAndDelete(temp_vector_m1,current_solution,Np + 2*Nr + 3*Ng);
                common_functions::copyAndDelete(temp_vector_m2,temp_vector_m1,Np + 2*Nr + 2*Ng);
                common_functions::copyAndDelete(temp_vector_m3,temp_vector_m2,Np + 2*Nr + Ng);
                if (dyndim_lifo)
                {
                    common_functions::copyAndDelete(temp_vector_m1,current_solution,Np + 2*Nr + 3*Ng);
                    common_functions::copyAndDelete(temp_vector_m2,temp_vector_m1,Np + 2*Nr + 2*Ng);
                    common_functions::copyAndDelete(temp_vector_m3,temp_vector_m2,Np + 2*Nr + Ng);
                }
                else
                {
                    remove_index_1 = common_functions::intUnifRand(Np + 2*Nr + 1,Np + 2*Nr + Ng); 
                    remove_index_2 = common_functions::intUnifRand(Np + 2*Nr + Ng + 1,Np + 2*Nr + 2*Ng); 
                    remove_index_3 = common_functions::intUnifRand(Np + 2*Nr + 2*Ng + 1,Np + 2*Nr + 3*Ng); 
                    common_functions::copyAndDelete(temp_vector_m1,current_solution,remove_index_3);
                    common_functions::copyAndDelete(temp_vector_m1,current_solution,remove_index_2);
                    common_functions::copyAndDelete(temp_vector_m1,current_solution,remove_index_1);
                }

                proposed_solution = temp_vector_m3;
            }
            proposed_Np = Np_new;
            proposed_Nr = Nr_new;
            proposed_Ng = Ng_new;

        }
        else
        {
            double p;
            double a,b,c;
            double step_length;
            if (update_single_parameter)
            {
                int update_index = static_cast<int>(floor((no_parameters)*rand01_1));
                category_index = common_functions::categoryIndex(update_index,Np,Nr,Ng); 
                proposed_solution = current_solution;
                p = common_functions::sgn_double(-1.0 + 2.0*rand01_2);
                if ((poly_term_log_sample == 1) && (category_index == 1))
                {
                    cout << "hei" << endl;
                    a = log(min_step_length_values(category_index));
                    c_min = common_functions::minPolynomialValue(some_model,update_index);
                    c_max = common_functions::maxPolynomialValue(some_model,update_index);
                    if (p < 0)
                        c = log(current_solution(update_index) - c_min);
                    else
                        c = log(c_max - current_solution(update_index));
                    if (c < log(min_step_length_values(category_index)))
                        p = -1.0*p;
                    if (p < 0)
                        c = log(current_solution(update_index) - c_min);
                    else
                        c = log(c_max - current_solution(update_index));
                    b = std::min(c,log(max_step_length_values(category_index)));
                    step_length = a + (b - a)*rand01_3;
                    proposed_solution(update_index) = current_solution(update_index) + p*exp(step_length);
                }
                else
                {
                    a = min_step_length_values(category_index);
                    if (p < 0)
                        c = current_solution(update_index) - min_parameter_values(category_index);
                    else
                        c = max_parameter_values(category_index) - current_solution(update_index);
                    if (c < min_step_length_values(category_index))
                        p = -1.0*p;
                    if (p < 0)
                        c = current_solution(update_index) - min_parameter_values(category_index);
                    else
                        c = max_parameter_values(category_index) - current_solution(update_index);
                    b = std::min(c,max_step_length_values(category_index));
                    step_length = a + (b - a)*rand01_4;
                    proposed_solution(update_index) = current_solution(update_index) + p*step_length;
                }
            }
            else
            {
                for (i = 0; i < no_parameters; ++i)
                {
                    rand01_5 = (double) rand() / RAND_MAX;
                    rand01_6 = (double) rand() / RAND_MAX;
                    rand01_7 = (double) rand() / RAND_MAX;
                
                    p = common_functions::sgn_double(-1.0 + 2.0*rand01_5);
                    category_index = common_functions::categoryIndex(i,Np,Nr,Ng); 
                    c_min = common_functions::minPolynomialValue(some_model,i);
                    c_max = common_functions::maxPolynomialValue(some_model,i);
                    if ((poly_term_log_sample == 1) && (category_index == 1))
                    {
                        a = log(min_step_length_values(category_index));
                        if (p < 0)
                            c = log(current_solution(i) - c_min);
                        else
                            c = log(c_max - current_solution(i));
                        if (c < log(min_step_length_values(category_index)))
                            p = -1.0*p;
                        if (p < 0)
                            c = log(current_solution(i) - c_min);
                        else
                            c = log(c_max - current_solution(i));
                        b = std::min(c,log(max_step_length_values(category_index)));
                        step_length = a + (b - a)*rand01_6;
                        proposed_solution(i) = current_solution(i) + p*exp(step_length);
                    }
                    else
                    {
                        a = min_step_length_values(category_index);
                        if (p < 0)
                            c = current_solution(i) - min_parameter_values(category_index);
                        else
                            c = max_parameter_values(category_index) - current_solution(i);
                        if (c < min_step_length_values(category_index))
                            p = -1.0*p;
                        if (p < 0)
                            c = current_solution(i) - min_parameter_values(category_index);
                        else
                            c = max_parameter_values(category_index) - current_solution(i);
                        b = std::min(c,max_step_length_values(category_index));
                        step_length = a + (b - a)*rand01_7;
                        proposed_solution(i) = current_solution(i) + p*step_length;
                    }
                }
            }
        
        }
        proposed_Np = Np_new;
        proposed_Nr = Nr_new;
        proposed_Ng = Ng_new;
    }
    else
    {

        if (new_solution_method == 1)  // Old way 
        {
            Eigen::VectorXd direction = Eigen::MatrixXd::Zero(no_parameters,1);
            newDirection(some_model, direction, min_parameter_values, max_parameter_values, latest_solution_values, latest_misfit_values, global_iterator, Np, Nr, Ng);
        
            /*
            for (i = 0; i < no_parameters; ++i)
            {
                if (current_solution(i) - min_parameter_values(i) < 1e-8)
                {    
                    current_solution(i) = min_parameter_values(i);
                    if (direction(i) < 0.0)
                        direction(i) = 0.0;
                }
                if (max_parameter_values(i) - current_solution(i) < 1e-8)
                {
                    current_solution(i) = max_parameter_values(i);
                    if (direction(i) > 0.0)
                        direction(i) = 0.0;
                }
            }
            */
        
            double step_length = stepLength(some_model, direction, current_solution, min_parameter_values, max_parameter_values, num_halves, Np, Nr, Ng);

            proposed_solution = current_solution + step_length*direction;
        }
        else if (new_solution_method == 2) // Individual step length and/or block (single) sampler
        {
            double p;
            double a,b,c;
            double step_length;
            if (update_single_parameter)
            {
                int update_index = static_cast<int>(floor((no_parameters)*rand01_1));
                category_index = common_functions::categoryIndex(update_index,Np,Nr,Ng);
                proposed_solution = current_solution;
                p = common_functions::sgn_double(-1.0 + 2.0*rand01_2);
                if ((poly_term_log_sample == 1) && (category_index == 1))
                {
                    a = log(min_step_length_values(category_index));
                    c_min = common_functions::minPolynomialValue(some_model,update_index);
                    c_max = common_functions::maxPolynomialValue(some_model,update_index);
                    if (p < 0)
                        c = log(current_solution(update_index) - c_min);
                    else
                        c = log(c_max - current_solution(update_index));
                    if (c < log(min_step_length_values(category_index)))
                        p = -1.0*p;
                    if (p < 0)
                        c = log(current_solution(update_index) - c_min);
                    else
                        c = log(c_max - current_solution(update_index));
                    b = std::min(c,log(max_step_length_values(category_index)));
                    step_length = a + (b - a)*rand01_3;
                    proposed_solution(update_index) = current_solution(update_index) + p*exp(step_length);
                }
                else
                {
                    a = min_step_length_values(category_index);
                    if (p < 0)
                        c = current_solution(update_index) - min_parameter_values(category_index);
                    else
                        c = max_parameter_values(category_index) - current_solution(update_index);
                    if (c < min_step_length_values(category_index))
                        p = -1.0*p;
                    if (p < 0)
                        c = current_solution(update_index) - min_parameter_values(category_index);
                    else
                        c = max_parameter_values(category_index) - current_solution(update_index);
                    b = std::min(c,max_step_length_values(category_index));
                    step_length = a + (b - a)*rand01_4;
                    proposed_solution(update_index) = current_solution(update_index) + p*step_length;
                }
            }
            else
            {
                for (i = 0; i < no_parameters; ++i)
                {
                    rand01_5 = (double) rand() / RAND_MAX;
                    rand01_6 = (double) rand() / RAND_MAX;
                    rand01_7 = (double) rand() / RAND_MAX;
                    
                    category_index = common_functions::categoryIndex(i,Np,Nr,Ng);
                    p = common_functions::sgn_double(-1.0 + 2.0*rand01_5);
                
                    if ((poly_term_log_sample == 1) && (category_index == 1))
                    {
                        c_min = common_functions::minPolynomialValue(some_model,i);
                        c_max = common_functions::maxPolynomialValue(some_model,i);
                        a = log(min_step_length_values(category_index));
                        if (p < 0)
                            c = log(current_solution(i) - c_min);
                        else
                            c = log(c_max - current_solution(i));
                        if (c < log(min_step_length_values(category_index)))
                            p = -1.0*p;
                        if (p < 0)
                            c = log(current_solution(i) - c_min);
                        else
                            c = log(c_max - current_solution(i));
                        b = std::min(c,log(max_step_length_values(category_index)));
                        step_length = a + (b - a)*rand01_6;
                        proposed_solution(i) = current_solution(i) + p*exp(step_length);
                    }
                    else
                    {
                        a = min_step_length_values(category_index);
                        if (p < 0)
                            c = current_solution(i) - min_parameter_values(category_index);
                        else
                            c = max_parameter_values(category_index) - current_solution(i);
                        if (c < min_step_length_values(category_index))
                            p = -1.0*p;
                        if (p < 0)
                            c = current_solution(i) - min_parameter_values(category_index);
                        else
                            c = max_parameter_values(category_index) - current_solution(i);
                        b = std::min(c,max_step_length_values(category_index));
                        step_length = a + (b - a)*rand01_7;
                        proposed_solution(i) = current_solution(i) + p*step_length;
                    }
                }
            }

        }
        else if (new_solution_method == 3) // EMGS
        {
            double a,b;
            double proposed_parameter_value; 
            if (update_single_parameter)
            {
                int update_index = static_cast<int>(floor((no_parameters)*rand01_1));
                category_index = common_functions::categoryIndex(update_index,Np,Nr,Ng);
                proposed_solution = current_solution;
                bool new_sample = true;
                double diff;
                while (new_sample)
                {
                    rand01_1 = (double) rand() / RAND_MAX;
                    proposed_parameter_value = min_parameter_values(category_index) + (max_parameter_values(category_index) - min_parameter_values(category_index))*rand01_1;
                    diff = abs(proposed_parameter_value - current_solution(category_index));
                    if ((diff <= min_step_length_values(category_index)) || (diff >= max_step_length_values(category_index)))
                        new_sample = true;
                    else
                        new_sample = false;
                }   
            
                proposed_solution(update_index) = proposed_parameter_value;
            }
        }
    }
}

void SimulatedAnnealing::newDirection(model1D::Model1D some_model, Ref<Eigen::VectorXd> direction, const Ref<Eigen::VectorXd>& min_parameter_values, const Ref<Eigen::VectorXd>& max_parameter_values, const Ref<Eigen::MatrixXd>& latest_solution_values, const Ref<Eigen::VectorXd>& latest_misfit_values, int global_iter_counter, int Np, int Nr, int Ng)
{
    if ((direction_type == 1) || (global_iter_counter <= no_iter_history))
    {
        uniformDirectionSampler(some_model, direction, min_parameter_values, max_parameter_values, Np, Nr, Ng);
    }
    
    if ((direction_type == 2) && (global_iter_counter > no_iter_history))
    {
        if (latest_misfit_values(no_iter_history - 1) == latest_misfit_values(no_iter_history - 2))
            uniformDirectionSampler(some_model, direction, min_parameter_values, max_parameter_values, Np, Nr, Ng);
        else
            adaptedGaussianDirectionSampler(some_model, direction, latest_solution_values, latest_misfit_values, min_parameter_values, max_parameter_values, Np, Nr, Ng);
    }
}

void SimulatedAnnealing::uniformDirectionSampler(model1D::Model1D some_model, Ref<Eigen::VectorXd> direction, const Ref<Eigen::VectorXd>& min_parameter_values, const Ref<Eigen::VectorXd>& max_parameter_values, int Np, int Nr, int Ng)
{
    double a;
    double b;
    double l_p = 0;
    double s_p = 0;
    double rand01, rand01_p, rand01_s;
    int category_index;
    for (int i = 0; i < direction.size(); ++i)
    {
        category_index = common_functions::categoryIndex(i,Np,Nr,Ng);
        if ((poly_term_log_sample == 1) && (category_index == 1))
        {
            rand01_p = ((double)rand()/(double)RAND_MAX);
            rand01_s = ((double)rand()/(double)RAND_MAX);
            a = -21.0;
            b = log(max_parameter_values(category_index) - min_parameter_values(category_index));
            l_p = a + (b - a)*rand01_p;
            s_p = common_functions::sgn_double(-1.0 + 2.0*rand01_s);
            direction(i) = s_p*exp(l_p);
        }
        else
        {
            rand01 = ((double)rand()/(double)RAND_MAX);
            a = min_parameter_values(category_index) - max_parameter_values(category_index);
            b = max_parameter_values(category_index) - min_parameter_values(category_index);
            direction(i) = a + (b - a)*rand01;
        }
    }
    //double euclidean_norm_direction = direction.norm();
    //for (int j = 0; j < direction.size(); ++j)
    //    direction(j) = direction(j)/euclidean_norm_direction;
}


void SimulatedAnnealing::adaptedGaussianDirectionSampler(model1D::Model1D some_model, Ref<Eigen::VectorXd> direction, const Ref<Eigen::MatrixXd>& latest_solution_values, const Ref<Eigen::VectorXd>& latest_misfit_values, const Ref<Eigen::VectorXd>& min_parameter_values, const Ref<Eigen::VectorXd>& max_parameter_values, int Np, int Nr, int Ng)
{
    // latest_solution_values is of size no_iter_history x no_parameters
    // latest_misfit_values is of size no_iter_history x 1
    // both could consisto of both accepted and rejected values

    int i,j,k;
    int no_parameters = some_model.get_no_parameters();

    // Make memory decending function gamma

    Eigen::VectorXd temp_x(no_iter_history);
    temp_x.setLinSpaced(no_iter_history,0,no_iter_history-1);
    Eigen::VectorXd gamma(temp_x.size());
    for (i = 0; i < temp_x.size(); ++i)
        gamma(i) = temp_x(i)*temp_x(i)*temp_x(i);
    for (i = 0; i < temp_x.size(); ++i)
        gamma(i) = gamma(i)/gamma.maxCoeff();

    // Make correlation vector
    Eigen::VectorXd correlation_vector = Eigen::MatrixXd::Zero(no_parameters,1);
    Eigen::VectorXd temp_col(no_iter_history);
    for (i = 0; i < no_parameters; ++i)
    {   
        for (j = 0; j < no_iter_history; ++j)
            temp_col(j) = latest_solution_values(j,i);
        correlation_vector(i) = common_functions::correlation(temp_col,latest_misfit_values);
    }

    // Make difference vectors
    Eigen::MatrixXd diff_solutions = Eigen::MatrixXd::Zero(no_iter_history-1, no_parameters);
    Eigen::MatrixXd diff_misfit    = Eigen::MatrixXd::Zero(no_iter_history-1,1);

    for (k = 0; k < no_iter_history-1; ++k)
    {
        diff_misfit(k) = latest_misfit_values(k+1) - latest_misfit_values(k);
        for (i = 0; i < no_parameters; ++i)
            diff_solutions(k,i) = latest_solution_values(k+1,i) - latest_solution_values(k,i);
    }
    
    // Make weights
    Eigen::MatrixXd weights = Eigen::MatrixXd::Zero(no_iter_history-1,no_parameters);

    for (i = 0; i < no_parameters; ++i)
    {
        for (k = 0; k < no_iter_history-1; ++k)
        {
            weights(k,i) = gamma(k)*fabs(diff_misfit(k))*fabs(correlation_vector(i));
        }
    }

    // Make weighted mean
    Eigen::VectorXd weighted_mean = Eigen::MatrixXd::Zero(no_parameters,1);
    double sum_weights_i;

    for (i = 0; i < no_parameters; ++i)
    {
        sum_weights_i = weights.col(i).sum();
        if (sum_weights_i == 0.0)
        {
            sum_weights_i = (double) no_parameters;
        }
        for (k = 0; k < no_iter_history-1; ++k)
        {
            weighted_mean(i) += weights(k,i)*common_functions::sgn_double(diff_misfit(k))*(-1.0)*diff_solutions(k,i);
        }
        weighted_mean(i) = weighted_mean(i)/sum_weights_i;
    }

    // Make weighted variance (covariance matrix)
    Eigen::VectorXd weighted_variance = Eigen::MatrixXd::Zero(no_parameters,1);
    double temp_val;

    for (i = 0; i < no_parameters; ++i)
    {
        sum_weights_i = weights.col(i).sum();
        for (k = 0; k < no_iter_history-1; ++k)
        {
            temp_val = common_functions::sgn_double(diff_misfit(k))*(-1.0)*diff_solutions(k,i) - weighted_mean(i);
            weighted_variance(i) += weights(k,i)*temp_val*temp_val;
        }
        if (weighted_variance(i) == 0)
        {
            cout << "Variance = 0 in adaptedGaussianDirectionSampler" << endl;
            break;
        }
        weighted_variance(i) = weighted_variance(i)/sum_weights_i;
    }
    int category_index;
    Eigen::MatrixXd weighted_covariance_matrix = Eigen::MatrixXd::Zero(no_parameters,no_parameters);
    for (i = 0; i < no_parameters; ++i)
    {
        category_index = common_functions::categoryIndex(i,Np,Nr,Ng);
        //weighted_covariance_matrix(i,i) = weighted_variance(i);
        weighted_covariance_matrix(i,i) = max_parameter_values(category_index) - min_parameter_values(category_index);
    }

    // Sample direction
    common_functions::multivariateGaussian(direction,weighted_mean,weighted_covariance_matrix); 
}

double SimulatedAnnealing::stepLength(model1D::Model1D some_model, const Ref<const Eigen::VectorXd> direction, const Ref<const Eigen::VectorXd> current_parameter_values, const Ref<Eigen::VectorXd>& min_parameter_values, const Ref<Eigen::VectorXd>& max_parameter_values, int num_halves, int Np, int Nr, int Ng)
{
    Eigen::VectorXd temp_min = Eigen::MatrixXd::Zero(direction.size(),1); 
    Eigen::VectorXd temp_max = Eigen::MatrixXd::Zero(direction.size(),1); 
    double no_parameters = some_model.get_no_parameters();
    int i;
    double temp;
    double c_min;
    double c_max;
    int category_index;
    for (i = 0; i < no_parameters; i++)
    {
        category_index = common_functions::categoryIndex(i,Np,Nr,Ng);
        if (direction(i) == 0.0)
        {
            temp_min(i) = 0.0;
            temp_max(i) = 1000000.0;
        }
        else
        {   
            if (category_index == 1)
            {
                c_min = common_functions::minPolynomialValue(some_model,i);
                c_max = common_functions::maxPolynomialValue(some_model,i);
                temp_min(i) = (c_min - current_parameter_values(i))/direction(i); 
                temp_max(i) = (c_max - current_parameter_values(i))/direction(i);
            }
            else
            {
                temp_min(i) = (min_parameter_values(category_index) - current_parameter_values(i))/direction(i); 
                temp_max(i) = (max_parameter_values(category_index) - current_parameter_values(i))/direction(i);
            }
            if (direction(i) < 0)
            {
                temp = temp_min(i);
                temp_min(i) = temp_max(i);
                temp_max(i) = temp;
            }
        }
    }
    double lambda_max = temp_max.minCoeff();
    lambda_max = lambda_max*pow(2,-(double)num_halves);
    double rand01 = ((double)rand()/(double)RAND_MAX);
    double lambda;
    if (poly_term_log_sample == 2)
    {
        double l_lambda = -8.0 + (log(lambda_max) + 8.0)*rand01;
        lambda = pow(10.0,l_lambda);
    }
    else
    {
        lambda = lambda_max*rand01;
    }
    
    return lambda;
}

void SimulatedAnnealing::stepLengthReduction(Ref<Eigen::VectorXd> min_step_length, Ref<Eigen::VectorXd> max_step_length, double factor, int power)
{
    // Reduce step length with factor factor of power power

    for (int i = 0; i < min_step_length.size(); ++i)
    {
        min_step_length(i) = min_step_length(i)*pow(factor, - (double) power);
        max_step_length(i) = max_step_length(i)*pow(factor, - (double) power);
    }
}

bool SimulatedAnnealing::writeVariableValuesToTxtFile(model1D::Model1D some_model, const Ref<Eigen::VectorXd>& initial_parameter_values, const Ref<Eigen::VectorXd>& min_parameter_values, const Ref<Eigen::VectorXd>& max_parameter_values, std::string filepath, std::string initial_filepath)
{
    // Arguments from initial_filepath file
    std::string _name;
    std::string result_type;
    std::string solution_method;
    bool _noise;
    bool _dynamic_dimensions;
    int _no_poly_terms;
    int _no_resistor_terms;
    int _no_wavelet_terms;
    double _initial_temperature;
    double _zs;
    double _zr;
    double _x0;
    double _y0;
    bool _intow;
    double _r0;                          
    double _dr;
    double _rl;
    double _psi;
    double _tet;
    double _water_depth;
    double _z_formation_max;
    double _water_resistivity;
    double _resistor_thickness;
    double _min_misfit_value;
    bool _max_iter_termination;
    int _max_iter;
    bool _fn_value_termination;
    bool _reduce_step_length;
    int _epoch_const;
    int _num_iter_before_update;
    double _update_threshold;
    int _no_iter_history;
    int _direction_type;
    int _temperature_schedule;        // 1: romeijn, 2: geometric, 3: lundy
    bool _update_temp_while_accept;
    double _p_chi;            
    double _geometric_temp_const;
    double _lundy_temp_const;
    int _poly_term_log_sample;        // 1: logarithmic uniform direction sampler, 2: logarithmic uniform step length sampler
    int _new_solution_method;        // 1: old way, 2: new, 3: EMGS
    bool _update_single_parameter;
    bool _dyndim_lifo;
    int _min_Np;        // minimum order of polynomial term, 0 means just the constant term
    int _min_Nr;        // minimum number of resistors
    int _min_Ng;        // minimum number of gaussfunctions
    int _max_Np;       // maximum order of polynomial term
    int _max_Nr;       // maximum number of resistors
    int _max_Ng;       // maximum number of resistors
    double _p_nothing;      // probability of not changing number of parameters
    double _p_in_Np;      // probability of increase Np by one       
    double _p_de_Np;      // probability of decrease Np by one
    double _p_in_Nr;
    double _p_de_Nr;
    double _p_in_Ng;
    double _p_de_Ng;
    double _min_poly_contribution;
    double _max_poly_contribution;
    Eigen::VectorXd _min_step_length_values(7);
    Eigen::VectorXd _max_step_length_values(7);
    double _best_model_prob;
    Eigen::VectorXd _initial_category_values(7);
    Eigen::VectorXd _category_standard_deviation(7);
    int _initial_distribution;
    int _seed;


    std::ifstream initial_in_stream(initial_filepath.c_str(), std::ios::in);
    if (!initial_in_stream.is_open())
        std::cout << "Unable to open file()" << endl;
    initial_in_stream >> _name;
    initial_in_stream >> result_type;
    initial_in_stream >> solution_method;
    initial_in_stream >> _noise;
    initial_in_stream >> _dynamic_dimensions;
    initial_in_stream >> _no_poly_terms;
    initial_in_stream >> _no_resistor_terms;
    initial_in_stream >> _no_wavelet_terms;
    initial_in_stream >> _initial_temperature;
    initial_in_stream >> _zs;
    initial_in_stream >> _zr;
    initial_in_stream >> _x0;
    initial_in_stream >> _y0;
    initial_in_stream >> _intow;
    initial_in_stream >> _r0;
    initial_in_stream >> _dr;
    initial_in_stream >> _rl;
    initial_in_stream >> _psi;
    initial_in_stream >> _tet;
    initial_in_stream >> _water_depth;
    initial_in_stream >> _z_formation_max;
    initial_in_stream >> _water_resistivity;
    initial_in_stream >> _resistor_thickness;
    int i;
    int _no_parameters = _no_poly_terms + 2*_no_resistor_terms + 3*_no_wavelet_terms;
    Eigen::VectorXd _parameters(_no_parameters);
    for (i = 0; i < _no_parameters; ++i)
        initial_in_stream >> _parameters[i];
    initial_in_stream >> _min_misfit_value;
    initial_in_stream >> _max_iter_termination;
    initial_in_stream >> _max_iter;
    initial_in_stream >> _fn_value_termination;
    initial_in_stream >> _reduce_step_length;
    initial_in_stream >> _epoch_const;
    initial_in_stream >> _num_iter_before_update;
    initial_in_stream >> _update_threshold;
    initial_in_stream >> _no_iter_history;
    initial_in_stream >> _direction_type;
    initial_in_stream >> _temperature_schedule;
    initial_in_stream >> _update_temp_while_accept;
    initial_in_stream >> _p_chi;
    initial_in_stream >> _geometric_temp_const;
    initial_in_stream >> _lundy_temp_const;
    initial_in_stream >> _poly_term_log_sample;
    initial_in_stream >> _new_solution_method;
    initial_in_stream >> _update_single_parameter;
    initial_in_stream >> _dyndim_lifo;
    initial_in_stream >> _min_Np;
    initial_in_stream >> _min_Nr;
    initial_in_stream >> _min_Ng;
    initial_in_stream >> _max_Np;
    initial_in_stream >> _max_Nr;
    initial_in_stream >> _max_Ng;
    initial_in_stream >> _p_nothing;
    initial_in_stream >> _p_in_Np;
    initial_in_stream >> _p_de_Np;
    initial_in_stream >> _p_in_Nr;
    initial_in_stream >> _p_de_Nr;
    initial_in_stream >> _p_in_Ng;
    initial_in_stream >> _p_de_Ng;
    initial_in_stream >> _min_poly_contribution;
    initial_in_stream >> _max_poly_contribution;

    for (i = 0; i < 7; ++i)
        initial_in_stream >> _min_step_length_values[i];
    
    for (i = 0; i < 7; ++i)
        initial_in_stream >> _max_step_length_values[i];
    
    initial_in_stream >> _best_model_prob;

    for (i = 0; i < 7; ++i)
        initial_in_stream >> _initial_category_values[i];
    
    for (i = 0; i < 7; ++i)
        initial_in_stream >> _category_standard_deviation[i];

    initial_in_stream >> _initial_distribution;
    initial_in_stream >> _seed;
    initial_in_stream.close();

    std::ofstream file(filepath.c_str(), std::ios::out | std::ios::trunc);
    if (!file.is_open())
        return false;
    
    file << "Variable and parameter values for this particular run\n";
    file << "\n";
    
    file << "Model1D.h:\n";
    file << std::setw(45) << "name:    "                    << some_model.get_name() << "\n";
    file << std::setw(45) << "result_type:    "             << result_type << "\n";
    file << std::setw(45) << "solution_method:    "         << solution_method << "\n";
    file << std::setw(45) << "min_misfit_value:    "        << some_model.get_min_misfit_value() << "\n";
    file << std::setw(45) << "water_depth:    "             << some_model.get_water_depth() << "\n";
    file << std::setw(45) << "water_resistivity:    "       << some_model.get_water_resistivity() << "\n";
    file << std::setw(45) << "resistor_thickness:    "      << some_model.get_resistor_thickness() << "\n";
    file << std::setw(45) << "z_water_min:    "             << some_model.get_z_water_min() << "\n";
    file << std::setw(45) << "z_formation_max:    "         << some_model.get_z_formation_max() << "\n";
    file << std::setw(45) << "initial_z_spacing:    "       << some_model.get_initial_z_spacing() << "\n";
    file << std::setw(45) << "z_formation_vector_length:    " << some_model.get_z_formation_vector_length() << "\n";
    file << std::setw(45) << "par:    "                     << some_model.get_par() << "\n";
    file << std::setw(45) << "freq:    "                    << some_model.get_freq() << "\n";
    file << std::setw(45) << "mhor:    "                    << some_model.get_mhor() << "\n";
    file << std::setw(45) << "zs:    "                      << some_model.get_zs() << "\n";
    file << std::setw(45) << "zr:    "                      << some_model.get_zr() << "\n";
    file << std::setw(45) << "r0:    "                      << some_model.get_r0() << "\n";
    file << std::setw(45) << "dr:    "                      << some_model.get_dr() << "\n";
    file << std::setw(45) << "rl:    "                      << some_model.get_rl() << "\n";
    file << std::setw(45) << "sortyp:    "                  << some_model.get_sortyp() << "\n";
    file << std::setw(45) << "camp:    "                    << some_model.get_camp() << "\n";
    file << std::setw(45) << "alen:    "                    << some_model.get_alen() << "\n";
    file << std::setw(45) << "azet:    "                    << some_model.get_azet() << "\n";
    file << std::setw(45) << "tet:    "                     << some_model.get_tet() << "\n";
    file << std::setw(45) << "method:    "                  << some_model.get_method() << "\n";
    file << std::setw(45) << "acur:    "                    << some_model.get_acur() << "\n";
    file << std::setw(45) << "comp:    "                    << some_model.get_comp() << "\n";
    file << std::setw(45) << "gath:    "                    << some_model.get_gath() << "\n";
    file << std::setw(45) << "tvi:    "                     << some_model.get_tvi() << "\n";
    file << std::setw(45) << "psi:    "                     << some_model.get_psi() << "\n";
    file << std::setw(45) << "x0:    "                      << some_model.get_x0() << "\n";
    file << std::setw(45) << "y0:    "                      << some_model.get_y0() << "\n";
    file << std::setw(45) << "zlay_obs:    "                << some_model.get_zlay_obs() << "\n";
    file << std::setw(45) << "mver_obs:    "                << some_model.get_mver_obs() << "\n";
    file << std::setw(45) << "noise:    "                   << some_model.get_noise() << "\n";
    file << std::setw(45) << "alpha:    "                   << some_model.get_alpha() << "\n";
    file << std::setw(45) << "eta_squared:    "             << some_model.get_eta_squared() << "\n";
    file << std::setw(45) << "run_id:    "                  << some_model.get_run_id() << "\n";
    file << std::setw(45) << "no_receivers:    "            << some_model.get_no_receivers() << "\n";
    file << std::setw(45) << "no_sources:    "              << some_model.get_no_sources() << "\n";
    file << std::setw(45) << "no_frequencies:    "          << some_model.get_no_frequencies() << "\n";
    file << std::setw(45) << "no_field_components:    "     << some_model.get_no_field_components() << "\n";
    file << std::setw(45) << "no_data:    "                 << some_model.get_no_data() << "\n";
    file << std::setw(45) << "no_poly_terms:    "           << some_model.get_no_poly_terms() << "\n";
    file << std::setw(45) << "no_resistor_terms:    "       << some_model.get_no_resistor_terms() << "\n";
    file << std::setw(45) << "no_wavelet_terms:    "        << some_model.get_no_wavelet_terms() << "\n";
    file << std::setw(45) << "no_parameters:    "           << some_model.get_no_parameters() << "\n";
    file << std::setw(45) << "log_sum:    "                 << some_model.get_log_sum() << "\n";
    file << "\n";

    file << "SimulatedAnnealing.h:\n";
    file << std::setw(45) << "dynamic dimensions:    "      << _dynamic_dimensions << "\n";
    file << std::setw(45) << "max_iter_termination:    "    << max_iter_termination << "\n";
    file << std::setw(45) << "max_iter:    "                << max_iter << "\n";
    file << std::setw(45) << "fn_value_termination:    "    << fn_value_termination << "\n";
    file << std::setw(45) << "reduce_step_length:    "      << reduce_step_length << "\n";
    file << std::setw(45) << "epoch constant:    "          << _epoch_const << "\n";
    file << std::setw(45) << "num_iter_before_update:    "  << num_iter_before_update << "\n";
    file << std::setw(45) << "update_threshold:    "        << update_threshold << "\n";
    file << std::setw(45) << "no_iter_history:    "         << no_iter_history << "\n";
    file << std::setw(45) << "direction_type:    "          << direction_type << "\n";
    file << std::setw(45) << "Initial temperature:    "     << _initial_temperature << "\n";
    file << std::setw(45) << "temperature_schedule:    "    << temperature_schedule << "\n";
    file << std::setw(45) << "update_temp_while_accept:    "<< _update_temp_while_accept << "\n";
    file << std::setw(45) << "p_chi:    "                   << p_chi << "\n";
    file << std::setw(45) << "geometric_temp_const:    "    << geometric_temp_const << "\n";
    file << std::setw(45) << "lundy_temp_const:    "        << lundy_temp_const << "\n";
    file << std::setw(45) << "poly_term_log_sample:    "    << poly_term_log_sample << "\n";
    file << std::setw(45) << "new_solution_method:    "     << new_solution_method << "\n";
    file << std::setw(45) << "update_single_parameter:    " << update_single_parameter << "\n";
    file << std::setw(45) << "dyndim_lifo:    "             << dyndim_lifo << "\n";
    file << std::setw(45) << "min_Np:    "                  << _min_Np << "\n";
    file << std::setw(45) << "max_Np:    "                  << _max_Np << "\n";
    file << std::setw(45) << "min_Nr:    "                  << _min_Nr << "\n";
    file << std::setw(45) << "max_Nr:    "                  << _max_Nr << "\n";
    file << std::setw(45) << "min_Ng:    "                  << _min_Ng << "\n";
    file << std::setw(45) << "max_Ng:    "                  << _max_Ng << "\n";
    file << std::setw(45) << "p_nothing:    "               << _p_nothing << "\n";
    file << std::setw(45) << "p_in_Np:    "                 << _p_in_Np << "\n";
    file << std::setw(45) << "p_de_Np:    "                 << _p_de_Np << "\n";
    file << std::setw(45) << "p_in_Nr:    "                 << _p_in_Nr << "\n";
    file << std::setw(45) << "p_de_Nr:    "                 << _p_de_Nr << "\n";
    file << std::setw(45) << "p_in_Ng:    "                 << _p_in_Ng << "\n";
    file << std::setw(45) << "p_de_Ng:    "                 << _p_de_Ng << "\n";
    file << std::setw(45) << "best model prob:    "         << _best_model_prob << "\n";
    file << std::setw(45) << "initial distribution:    "    << _initial_distribution << "\n";
    file << std::setw(45) << "Random seed:    "             << _seed << "\n";
    file << "\n";

    file << "Other data:\n";
    
    file << "\n";
    file << std::setw(45) << "Min parameter values:    \n";
    for (i = 0; i < 7; ++i)
        file << std::setw(45) << "    " << min_parameter_values[i] << "\n";
    
    file << "\n";
    file << std::setw(45) << "Max parameter values:    \n";
    for (i = 0; i < 7; ++i)
        file << std::setw(45) << "    " << max_parameter_values[i] << "\n";

    file << "\n";
    file << std::setw(45) << "Min step length values:    \n";
    for (i = 0; i < 7; ++i)
        file << std::setw(45) << "    " << _min_step_length_values[i] << "\n";
    
    file << "\n";
    file << std::setw(45) << "Max step length values:    \n";
    for (i = 0; i < 7; ++i)
        file << std::setw(45) << "    " << _max_step_length_values[i] << "\n";
    
    file << "\n";
    file << std::setw(45) << "Initial category values:    \n";
    for (i = 0; i < 7; ++i)
        file << std::setw(45) << "    " << _initial_category_values[i] << "\n";
    
    file << "\n";
    file << std::setw(45) << "Category standard deviations:    \n";
    for (i = 0; i < 7; ++i)
        file << std::setw(45) << "    " << _category_standard_deviation[i] << "\n";

    file << "\n";
    file << std::setw(45) << "Initial parameter values:    \n";
    for (int i = 0; i < initial_parameter_values.size(); ++i)
        file << std::setw(45) << "    " <<  initial_parameter_values(i) << "\n";

    file.close();
}
}
