#ifndef SIMULATED_ANNEALING_H
#define SIMULATED_ANNEALING_H

#include <vector>

#include <Eigen/Dense>
#include <Eigen/Core>
#include <Eigen/SVD>

#include "model1D.h"

using Eigen::Ref;

namespace simulated_annealing
{

class SimulatedAnnealing
{
    
  public:
    // Constructors:
    SimulatedAnnealing();
    SimulatedAnnealing(double p_chi_, bool max_iter_termination, int max_iter_, bool fn_value_termination_, bool reduce_step_length_, int num_iter_before_update_, double update_threshold_);
    
    // Regular functions:
    double updateTemperature(model1D::Model1D some_model, std::vector<double> misfit_values, double min_misfit_value, int no_parameters, double current_temperature, double initial_temperature, double global_iterate); 
    void newProposedSolution(model1D::Model1D some_model, Ref<Eigen::VectorXd> proposed_solution, const Ref<Eigen::VectorXd>& current_solution, const Ref<Eigen::VectorXd>& min_parameter_values, const Ref<Eigen::VectorXd>& max_parameter_values, const Ref<Eigen::MatrixXd>& latest_solution_values, const Ref<Eigen::VectorXd>& latest_misfit_values, int num_halves, int global_iterator, int& proposed_Np, int& proposed_Nr, int& proposed_Ng, int Np, int Nr, int Ng, const Ref<Eigen::VectorXd>& min_step_length_values, const Ref<Eigen::VectorXd>& max_step_length_values, const Ref<Eigen::VectorXd>& initial_category_values, const Ref<Eigen::VectorXd>& category_standard_deviation, int initial_distribution);
    void newDirection(model1D::Model1D some_model, Ref<Eigen::VectorXd> direction, const Ref<Eigen::VectorXd>& min_parameter_values, const Ref<Eigen::VectorXd>& max_parameter_values, const Ref<Eigen::MatrixXd>& latest_solution_values, const Ref<Eigen::VectorXd>& latest_misfit_values, int global_iter_counter, int Np, int Nr, int Ng);
    void uniformDirectionSampler(model1D::Model1D some_model, Ref<Eigen::VectorXd> direction, const Ref<Eigen::VectorXd>& min_parameter_values, const Ref<Eigen::VectorXd>& max_parameter_values, int Np, int Nr, int Ng);
    void adaptedGaussianDirectionSampler(model1D::Model1D some_model, Ref<Eigen::VectorXd> direction, const Ref<Eigen::MatrixXd>& latest_solution_values, const Ref<Eigen::VectorXd>& latest_misfit_values, const Ref<Eigen::VectorXd>& min_parameter_values, const Ref<Eigen::VectorXd>& max_parameter_values, int Np, int Nr, int Ng);
    double stepLength(model1D::Model1D some_model, const Ref<const Eigen::VectorXd> direction, const Ref<const Eigen::VectorXd> current_parameter_values, const Ref<Eigen::VectorXd>& min_parameter_values, const Ref<Eigen::VectorXd>& max_parameter_values, int num_halves, int Np, int Nr, int Ng);
    void stepLengthReduction(Ref<Eigen::VectorXd> min_step_length, Ref<Eigen::VectorXd> max_step_length, double factor, int power);
    bool writeVariableValuesToTxtFile(model1D::Model1D some_model, const Ref<Eigen::VectorXd>& initial_parameter_values, const Ref<Eigen::VectorXd>& min_parameter_values, const Ref<Eigen::VectorXd>& max_parameter_values, std::string filepath, std::string initial_filepath);

    // Accessor and mutator functions (set/get):
    
    void set_dynamic_dimensions(bool dynamic_dimensions_) {dynamic_dimensions = dynamic_dimensions_;}
    void set_max_iter_termination(bool max_iter_termination_) {max_iter_termination = max_iter_termination_;}
    void set_max_iter(int max_iter_) {max_iter = max_iter_;}
    void set_fn_value_termination(bool fn_value_termination_) {fn_value_termination = fn_value_termination_;}
    void set_reduce_step_length(bool reduce_step_length_) {reduce_step_length = reduce_step_length_;}
    void set_epoch_const(int epoch_const_) {epoch_const = epoch_const_;}
    void set_num_iter_before_update(int num_iter_before_update_) {num_iter_before_update = num_iter_before_update_;}
    void set_update_threshold(double update_threshold_) {update_threshold = update_threshold_;}
    void set_no_iter_history(int no_iter_history_) {no_iter_history = no_iter_history_;}
    void set_direction_type(int direction_type_) {direction_type = direction_type_;}
    void set_temperature_schedule(int temperature_schedule_) {temperature_schedule = temperature_schedule_;}
    void set_update_temp_while_accept(bool update_temp_while_accept_) {update_temp_while_accept = update_temp_while_accept_;}
    void set_p_chi (double p_chi_) {p_chi = p_chi_;}
    void set_geometric_temp_const(double geometric_temp_const_) {geometric_temp_const = geometric_temp_const_;}
    void set_lundy_temp_const(double lundy_temp_const_) {lundy_temp_const = lundy_temp_const_;}
    void set_poly_term_log_sample (int poly_term_log_sample_) {poly_term_log_sample = poly_term_log_sample_;}
    void set_new_solution_method(int new_solution_method_) {new_solution_method = new_solution_method_;}
    void set_update_single_parameter(bool update_single_parameter_) {update_single_parameter = update_single_parameter_;}
    void set_min_Np(int min_Np_) {min_Np = min_Np_;}
    void set_min_Nr(int min_Nr_) {min_Nr = min_Nr_;}
    void set_min_Ng(int min_Ng_) {min_Ng = min_Ng_;}
    void set_max_Np(int max_Np_) {max_Np = max_Np_;}
    void set_max_Nr(int max_Nr_) {max_Nr = max_Nr_;}
    void set_max_Ng(int max_Ng_) {max_Ng = max_Ng_;}
    void set_p_nothing(double p_nothing_) {p_nothing = p_nothing_;}
    void set_p_in_Np(double p_in_Np_) {p_in_Np = p_in_Np_;}
    void set_p_de_Np(double p_de_Np_) {p_de_Np = p_de_Np_;}
    void set_p_in_Nr(double p_in_Nr_) {p_in_Nr = p_in_Nr_;}
    void set_p_de_Nr(double p_de_Nr_) {p_de_Nr = p_de_Nr_;}
    void set_p_in_Ng(double p_in_Ng_) {p_in_Ng = p_in_Ng_;}
    void set_p_de_Ng(double p_de_Ng_) {p_de_Ng = p_de_Ng_;}
    void set_no_change(bool no_change_) {no_change = no_change_;}
    void set_increase_Np(bool increase_Np_) {increase_Np = increase_Np_;}
    void set_decrease_Np(bool decrease_Np_) {decrease_Np = decrease_Np_;}
    void set_increase_Nr(bool increase_Nr_) {increase_Nr = increase_Nr_;}
    void set_decrease_Nr(bool decrease_Nr_) {decrease_Nr = decrease_Nr_;}
    void set_increase_Ng(bool increase_Ng_) {increase_Ng = increase_Ng_;}
    void set_decrease_Ng(bool decrease_Ng_) {decrease_Ng = decrease_Ng_;}
    void set_dyndim_lifo(bool dyndim_lifo_) {dyndim_lifo = dyndim_lifo_;}
    void set_best_model_prob(double best_model_prob_) {best_model_prob = best_model_prob_;}

    bool get_max_iter_termination() {return max_iter_termination;}
    int get_max_iter() {return max_iter;}
    bool get_fn_value_termination() {return fn_value_termination;}
    bool get_reduce_step_length() {return reduce_step_length;}
    int get_epoch_const() {return epoch_const;}
    int get_num_iter_before_update() {return num_iter_before_update;}
    double get_update_threshold() {return update_threshold;}
    int get_no_iter_history() {return no_iter_history;}
    int get_direction_type() {return direction_type;}
    double get_temperature_schedule() {return temperature_schedule;}
    bool get_update_temp_while_accept() {return update_temp_while_accept;}
    double get_p_chi() {return p_chi;}
    double get_geometric_temp_const() {return geometric_temp_const;}
    double get_lundy_temp_const() {return lundy_temp_const;}
    int get_poly_term_log_sample() {return poly_term_log_sample;}
    int get_new_solution_method() {return new_solution_method;}
    bool get_update_single_parameter() {return update_single_parameter;}
    int get_min_Np() {return min_Np;}                     // minimum order of polynomial term, 0 means just the constant term
    int get_min_Nr() {return min_Nr;}                     // minimum number of resistors
    int get_min_Ng() {return min_Ng;}                     // minimum number of gaussfunctions
    int get_max_Np() {return max_Np;}                     // maximum order of polynomial term
    int get_max_Nr() {return max_Nr;}                     // maximum number of resistors
    int get_max_Ng() {return max_Ng;}                     // maximum number of resistors
    double get_p_nothing() {return p_nothing;}            // probability of not changing number of parameters
    double get_p_in_Np() {return p_in_Np;}        
    double get_p_de_Np() {return p_de_Np;}
    double get_p_in_Nr() {return p_in_Nr;}
    double get_p_de_Nr() {return p_de_Nr;}
    double get_p_in_Ng() {return p_in_Ng;}
    double get_p_de_Ng() {return p_de_Ng;}
    bool get_dyndim_lifo() {return dyndim_lifo;}
    double get_best_model_prob() {return best_model_prob;}


  private:
    bool max_iter_termination;      // terminate the search when max_iter is reached
    int max_iter;                   // 
    bool fn_value_termination;      // terminate search if function value is below a fn_value_termination
    bool reduce_step_length;        // reduce max step length on/off
    int epoch_const;                // 
    int num_iter_before_update;     // number of iterations before one check if one should update (= epoch_const*number of variables)
    double update_threshold;        // if mean(fn_value_1, ... ,fn_value_num_iter_before_update) < update_threshold: reduce step length
    int no_iter_history;            // number of iterations to include in the adapted direction sampler
    int direction_type;             // Currently implemented types are 1: uniform, 2: adapted gaussian
    double temperature_schedule;    // 1:Romeijn, 2:Geometric, 3:Lundy
    bool update_temp_while_accept;  // Update the temperature if a new solution is accepted
    double p_chi;                   // 100(1-p) percentile in chi-squared distribution, used in Romeijn 
    double geometric_temp_const;    // used in geometric temperature schedule
    double lundy_temp_const;        // used in lundy temperature schedule
    int poly_term_log_sample;       // 1: logarithmic direction, 2: logarithmic step length 
    int new_solution_method;        // 1: old, 2: new, 3: EMGS
    bool update_single_parameter;
    bool dynamic_dimensions;        
    int min_Np;                     // minimum order of polynomial term, 0 means just the constant term
    int min_Nr;                     // minimum number of resistors
    int min_Ng;                     // minimum number of gaussfunctions
    int max_Np;                     // maximum order of polynomial term
    int max_Nr;                     // maximum number of resistors
    int max_Ng;                     // maximum number of resistors
    double p_nothing;               // probability of not changing number of parameters
    double p_in_Np;        
    double p_de_Np;
    double p_in_Nr;
    double p_de_Nr;
    double p_in_Ng;
    double p_de_Ng;
    bool no_change;
    bool increase_Np;
    bool decrease_Np;
    bool increase_Nr;
    bool decrease_Nr;
    bool increase_Ng;
    bool decrease_Ng;
    bool dyndim_lifo;

    // variables unique to the new transverse simulated annealing
    double best_model_prob;         // probability of chosing the model with best fit
};

}



#endif
